# coding=utf-8
# Copyright 2023-present the International Business Machines.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Our implementation of the VeriScore paper with open-source models (LLaMA)
# where facts/claims are checked with respect to contexts retrieved from an
# external source such as Wikipedia or Google Search (for the latter we consider
# the passage retrieved from the corresponding link).

import json
import asyncio
import time
import mellea.stdlib.functional as mfuncs

from typing import Any, Dict, List, Tuple
from mellea.backends import Backend
from mellea.stdlib.context import SimpleContext
from mellea.core import ModelOutputThunk
from mellea.stdlib.requirements import check
from mellea.stdlib.sampling import RejectionSamplingStrategy
from mellea.core import FancyLogger

# Local imports
from fact_reasoner.core.atomizer import Atomizer
from fact_reasoner.core.reviser import Reviser
from fact_reasoner.core.retriever import ContextRetriever
from fact_reasoner.core.base import Atom, Context
from fact_reasoner.core.utils import build_atoms, build_contexts, remove_duplicated_atoms
from fact_reasoner.utils import extract_last_square_brackets, LOOP_BUDGET

# Version 2 of the prompt (based on more recent work VeriScore, FactBench)
INSTRUCTION_VERISCORE = """

Instructions:
You are provided with a STATEMENT and several KNOWLEDGE points. \
Your task is to evaluate the relationship between the STATEMENT and the KNOWLEDGE, following the steps outlined below:

1. Summarize KNOWLEDGE Points: Carefully analyze the KNOWLEDGE points one by one and assess their relevance to the STATEMENT. \
Summarize the main points of the KNOWLEDGE.
2. Evaluate Evidence: Based on your reasoning:
- If the KNOWLEDGE strongly implies or directly supports the STATEMENT, explain the supporting evidence.
- If the KNOWLEDGE contradicts the STATEMENT, identify and explain the conflicting evidence.
- If the KNOWLEDGE is insufficient to confirm or deny the STATEMENT, explain why the evidence is inconclusive.
3. Restate the STATEMENT: After considering the evidence, restate the STATEMENT to maintain clarity.
4. Final Answer: Based on your reasoning and the STATEMENT, determine your final answer. \
Your final answer must be one of the following, wrapped in square brackets:
- [Supported] if the STATEMENT is supported by the KNOWLEDGE.
- [Contradicted] if the STATEMENT is contradicted by the KNOWLEDGE.
- [Unverifiable] if the KNOWLEDGE is insufficient to verify the STATEMENT.

Your task:

KNOWLEDGE: 
{{knowledge_text}}

STATEMENT:
{{atom_text}}
"""

class VeriScore:
    """
    Implementation of the VeriScore/FactBench paper. 

    Source:
        @misc{bayat2025factbench,
            title={FactBench: A Dynamic Benchmark for In-the-Wild Language Model Factuality Evaluation}, 
            author={Farima Fatahi Bayat and Lechen Zhang and Sheza Munir and Lu Wang},
            year={2025},
            eprint={2410.22257},
            archivePrefix={arXiv},
            primaryClass={cs.CL},
            url={https://arxiv.org/abs/2410.22257}, 
        }
    """

    def __init__(
            self,
            backend: Backend,
            atom_extractor: Atomizer = None,
            atom_reviser: Reviser = None,
            context_retriever: ContextRetriever = None,
    ):
        """
        Initialize the VeriScore pipeline.

        Args:
            backend: Backend
                The Mellea backend to use for LLM interactions.
            atom_extractor: Atomizer
                The atom decomposition component.
            atom_reviser: Reviser
                The atom reviser component.
            context_retriever: ContextRetriever
                The service used for retrieving external contexts.
        """

        self.backend = backend
        self.query = None
        self.response = None
        self.topic = None
        self.start_time = time.perf_counter() # get the start time

        self.context_retriever = context_retriever
        self.atom_extractor = atom_extractor
        self.atom_reviser = atom_reviser
        self.binary_output = False # default is False
    
        print(f"[VeriScore] Using Mellea backend: {self.backend.model_id}")
        print(f"[VeriScore] Binary output: {self.binary_output}")

        self.atoms = {} # indexed by atom id
        self.contexts = {} # indexed by context id

        # Ground truth labels (if any)
        self.labels_human = None
        
        # Disable Mellea logging
        FancyLogger.get_logger().setLevel(FancyLogger.ERROR)

    def from_dict_with_contexts(
            self,
            data: Dict[str, Any],
    ):
        """
        Initialize VeriScore from a dict containing both atoms and contexts.

        Args:
            data: Dict[str, Any]
                The dict containing the problem instance.
        """

        self.query = data["input"]
        self.response = data["output"]
        self.topic = data.get("topic", None)
        
        print(f"[VeriScore] Reading the atoms ...")                
        gold_labels = []
        atom_ids = []
        self.atoms = {}
        self.contexts = {}
        atom2contexts = {}
        for atom_dict in data["atoms"]:
            aid = atom_dict["id"]
            text = atom_dict["text"]
            original = atom_dict["original"]
            label = atom_dict.get("label", None)
            contexts = atom_dict["contexts"]
            a = Atom(id=aid, text=text, label=label)
            a.set_original(original)
            atom_ids.append(aid)
            gold_labels.append(label)
            self.atoms[aid] = a
            atom2contexts[aid] = contexts

        print(f"[VeriScore] Atoms found: {len(self.atoms)}")
        for _, atom in self.atoms.items():
            print(f"[VeriScore] {atom}")
        
        self.labels_human = dict(zip(atom_ids, gold_labels))
        print(f"[VeriScore] Lables found: {self.labels_human}")

        print(f"[VeriScore] Reading the contexts ...")
        for context_dict in data["contexts"]:
            cid = context_dict["id"]
            title = context_dict["title"]
            text = context_dict["text"]
            snippet = context_dict.get("snippet", "")
            link = context_dict.get("link", "")
            ctxt = Context(
                id=cid, 
                atom=None, 
                text=text, 
                title=title, 
                snippet=snippet, 
                link=link
            )
            self.contexts[cid] = ctxt

        print(f"[VeriScore] Contexts found: {len(self.contexts)}")
        for aid, atom in self.atoms.items():
            ctxts = []
            for c in atom2contexts[aid]:
                ctxts.append(self.contexts[c])
                self.contexts[c].set_atom(atom)
            atom.add_contexts(ctxts)
        
        print(f"[VeriScore] Pipeline initialized with {len(self.atoms)} atoms and {len(self.contexts)} contexts.")

    def to_json(self, json_file_path: str = None) -> Dict[str, Any]:
        """
        Save the VeriScore instance to a JSON file.

        Args:
            json_file: str
                The path to the output JSON file.
        """

        data = {}
        data["input"] = self.query
        data["output"] = self.response.strip()
        data["topic"] = self.topic
        data["atoms"] = []
        data["contexts"] = []

        # Write the atoms
        for aid, atom in self.atoms.items():
            atom_data = dict(
                id=aid, text=atom.get_text(), contexts=list(atom.get_contexts().keys())
            )
            if atom.get_label() is not None:
                atom_data["label"] = atom.get_label()
            data["atoms"].append(atom_data)

        # Write the contexts
        data["contexts"] = [context.to_json() for context in self.contexts.values()]

        # Write to a JSON file (if any)
        if json_file_path:
            with open(json_file_path, "w") as f:
                json.dump(data, f, indent=4)
            f.close()
            print(f"[VeriScore] Pipeline instance written to: {json_file_path}")

        return data

    def build(
            self,
            query: str = None,
            response: str = None,
            topic: str = None,
            has_atoms: bool = False,
            has_contexts: bool = False,
            revise_atoms: bool = False,
            use_fast_retriever: bool = True
    ):
        """
        Build the atoms and contexts using the retrieval service.

        Args:
            query: str
                The input user query.
            response: str
                The LLM generated response to the input query.
            topic: str
                The topic of the input query/response.
            has_atoms: bool
                A boolean flag indicating if the atoms have already been created.
            has_contexts: bool
                A boolean flag indicating if the contexts have already been created.
            revise_atoms: bool
                A boolean flag indicating that the atoms need to be decontextualized
                (i.e., pronouns he, she, it, ... replaced by the actual entity)
        """

        # Initialize the scorer
        if query is not None: 
            self.query = query
        if response is not None:
            self.response = response
        if topic is not None:
            self.topic = topic

        self.revise_atoms = revise_atoms

        # Safety checks
        assert self.atom_extractor is not None, \
            f"The atom extractor must be created."
        assert self.atom_reviser is not None, \
            f"The atom reviser must be created."

        print(f"[VeriScore] Building the pipeline ...")
        
        # Build the atoms 
        if has_atoms == False:
            self.atoms = build_atoms(
                response=self.response,
                atom_extractor=self.atom_extractor
            )
            self.revise_atoms = True # revise atoms if newly created
            print(f"[VeriScore] Extracted {len(self.atoms)} atoms.")
            for aid in self.atoms.keys():
                print(f"[VeriScore] {self.atoms[aid]}")

        assert len(self.atoms) > 0, \
            f"The atoms must be initialized before running the pipeline."

        # Revise the atoms
        if self.revise_atoms:
            print(f"[VeriScore] Revise the atoms ...")
            assert self.response is not None, f"The atom reviser requires a response."
            atom_ids = [aid for aid in sorted(self.atoms.keys())]
            old_atoms = [self.atoms[aid].get_text() for aid in atom_ids]
            result = asyncio.run(self.atom_reviser.run_batch(old_atoms, self.response))
            for i, aid in enumerate(atom_ids):
                elem = result[i]
                self.atoms[aid].set_text(elem["revised_unit"])
                print(f"[VeriScore] {self.atoms[aid]}")

        # Remove duplicated atoms (if any)
        self.atoms = remove_duplicated_atoms(self.atoms)
        print(f"[VeriScore] Created {len(self.atoms)} unique atoms.")

        # Build the contexts (per atom)
        if has_contexts == False: # check if contexts already in file
            self.contexts = build_contexts(
                atoms=self.atoms,
                retriever=self.context_retriever,
                use_fast_retriever=use_fast_retriever
            )

        print(f"[VeriScore] Retrieved {len(self.contexts)} contexts.")
        print(f"[VeriScore] Pipeline building completed.")

    def _get_label(self, output: ModelOutputThunk) -> str:
        """
        Extract the atom label from the generated text. We expect the label to
        be on the last line of the response, and be one of the following:
            [Supported], [Contradicted], [Unverifiable].
        We only consider [Supported]/S atoms, the others will be [NotSupported]/NS.
        """

        text = str(output)
        label = extract_last_square_brackets(text)
        if self.binary_output:
            if len(label) > 0 and label.lower() in ['supported']:
                return "S"
            else:
                return "NS"
        else:
            if len(label) > 0 and label.lower() in ['supported']:
                return "S"
            elif len(label) > 0 and label.lower() in ['contradicted']:
                return "C"
            else:
                return "U"
                
    async def predict_atom_labels(self) -> Tuple[Dict[str, str], Dict[str, str]]:
        """
        For each atom predict its label given the corresponding retrieved contexts.
        """

        # Safety checks
        assert len(self.atoms) > 0

        # Utility function to assemble the context of an atom
        def make_knowledge(passages: List[Dict[str, Any]]) -> str:
            knowledge = ""
            for _, psg in enumerate(passages):
                title = psg["title"]
                text = psg["text"]
                snippet = psg.get("snippet", "")
                knowledge += "Title: {}\nSummary: {}\nText: {}\n\n".format(title, snippet, text)
            
            return knowledge

        # Use the LLM to label the atom
        print(f"[VeriScore] Labeling atoms with {self.backend.model_id} ...")

        # Create the prompts for each of the atoms
        atom_ids = []
        atom_labels = []
        atom_outputs = []
        corutines = []
        for aid, atom in self.atoms.items():
            atom_ids.append(aid)
            atom_text = atom.get_text()
            contexts = atom.get_contexts()
            passages = []
            if contexts is not None and len(contexts) > 0:
                for _, c in contexts.items():
                    passages.append(dict(title=c.get_title(), text=c.get_text()))

            # prepare the context
            knowledge_text = make_knowledge(passages)
            print(f"[VeriScore] Processing atom: ({aid}) {atom_text}")

            # Execute the instruction
            corutine = mfuncs.ainstruct(
                INSTRUCTION_VERISCORE,
                context=SimpleContext(),
                backend=self.backend,
                requirements=[
                    check(
                        "The output must contain [Supported], [Contradicted] or [Unverifiable]"
                    )
                ],
                user_variables={"atom_text": atom_text, "knowledge_text": knowledge_text},
                strategy=RejectionSamplingStrategy(loop_budget=LOOP_BUDGET),
                return_sampling_results=True
            )
            corutines.append(corutine)

        print(f"[VeriScore] Awaiting for the async execution ...")
        outputs = await asyncio.gather(*(corutines[i] for i in range(len(corutines))))
        for output in outputs:
            label = self._get_label(output.result)
            atom_labels.append(label)
            atom_outputs.append(str(output))

        # Return the labeled atoms
        return dict(zip(atom_ids, atom_labels)), dict(zip(atom_ids, atom_outputs))
    
    def score(self) -> Dict[str, Any]:
        """
        Compute the factuality score taking into consideration the contexts 
        retrieved for each of the atom in the answer.

        Factuality score = # atoms(true) / # atoms

        Intuitively, a score of 100% means that all atoms in the answer are
        factually correct. If none of them are correct, then the score is 0%. If
        only half of the atoms are correct, then the score is 50%.

        Returns:
            Dict[str, Any]: The results dictionary containing the factuality score i.e., a real value in [0, 1]
        """

        # Run the VeriScore pipeline
        num_true_atoms = 0
        num_false_atoms = 0
        num_uniform_atoms = 0
        labels, raw_outputs = asyncio.run(self.predict_atom_labels())
        for _, label in labels.items():
            if self.binary_output:
                if label == "S":
                    num_true_atoms += 1
                else:
                    num_false_atoms += 1
            else:
                if label == "S":
                    num_true_atoms += 1
                elif label == "C":
                    num_false_atoms += 1
                else:
                    num_uniform_atoms += 1
      
        # Precision, R@K and F1@K
        fscore = float(num_true_atoms)/float(len(self.atoms))
        K = int(len(self.atoms) / 2) # K is assumed to be half
        recall_k = min(float(num_true_atoms)/K, 1.0)
        try:
            f1k = 2 * fscore * recall_k / (fscore + recall_k)
        except Exception as _:
            f1k = 0.0

        # Elapsed time
        elapsed_time = time.perf_counter() - self.start_time # elapsed time

        results = {}
        results["factuality_score"] = fscore
        results["recall_k"] = recall_k
        results["f1_k"] = f1k
        results["num_atoms"] = len(self.atoms)
        results["num_contexts"] = len(self.contexts)
        results["num_true_atoms"] = num_true_atoms
        results["num_false_atoms"] = num_false_atoms
        results["num_uniform_atoms"] = num_uniform_atoms
        results["entropy"] = None
        results["avg_entropy"] = None

        print(f"[VeriScore] Predictions: {labels}")
        if self.labels_human is not None and self.binary_output is True:
            true_atoms = 0
            false_atoms = 0
            num_true_positive = 0
            num_true_negative = 0
            num_false_positive = 0
            num_false_negative = 0
            for aid, l in self.labels_human.items():
                if l == "S":
                    true_atoms += 1
                    if labels[aid] == "S":
                        num_true_positive += 1
                    else:
                        num_false_negative += 1
                else:
                    false_atoms += 1
                    if labels[aid] == "NS":
                        num_true_negative += 1
                    else:
                        num_false_positive += 1                    
            fscore_gold = true_atoms/len(self.labels_human)
            print(f"[VeriScore] Gold labels: {self.labels_human}")
            print(f"[VeriScore] Predictions: {labels}")
            print(f"[VeriScore] Gold fscore: {fscore_gold} ({true_atoms}/{len(self.labels_human)})")
            results["gold_factuality_score"] = fscore_gold
            results["gold_true_atoms"] = true_atoms
            results["true_positive"] = num_true_positive
            results["true_negative"] = num_true_negative
            results["false_positive"] = num_false_positive
            results["false_negative"] = num_false_negative
        elif self.labels_human is not None and self.binary_output is False:
            true_atoms = 0
            false_atoms = 0
            num_true_positive = 0
            num_true_negative = 0
            num_false_positive = 0
            num_false_negative = 0
            for aid, l in self.labels_human.items(): # true labels are either S or NS
                if l == "S":
                    true_atoms += 1
                    if labels[aid] == "S":
                        num_true_positive += 1
                    else:
                        num_false_negative += 1
                else:
                    false_atoms += 1
                    if labels[aid] in ["C", "U"]:
                        num_true_negative += 1
                    else:
                        num_false_positive += 1     

            fscore_gold = true_atoms/len(self.labels_human)
            print(f"[VeriScore] Gold labels: {self.labels_human}")
            print(f"[VeriScore] Predictions: {labels}")
            print(f"[VeriScore] Gold fscore: {fscore_gold} ({true_atoms}/{len(self.labels_human)})")
            results["gold_factuality_score"] = fscore_gold
            results["gold_true_atoms"] = true_atoms
            results["true_positive"] = num_true_positive
            results["true_negative"] = num_true_negative
            results["false_positive"] = num_false_positive
            results["false_negative"] = num_false_negative

        results["topic"] = self.topic
        results["query"] = self.query
        results["response"] = self.response
        results["elapsed_time"] = elapsed_time
        results["predictions"] = labels
        results["raw_outputs"] = raw_outputs
        print(f"[VeriScore] Elapsed time: {elapsed_time:.4f} seconds.")

        return results




# coding=utf-8
# Copyright 2023-present the International Business Machines.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Context summarization using LLMs

import math
import asyncio
import mellea.stdlib.functional as mfuncs

from typing import Any, Dict, List
from mellea.backends import Backend
from mellea.backends import ModelOption
from mellea.stdlib.context import SimpleContext
from mellea.core import ModelOutputThunk
from mellea.stdlib.sampling import RejectionSamplingStrategy
from mellea.core import FancyLogger

from fact_reasoner.utils import LOOP_BUDGET

INSTRUCTION_WITHOUT_REF = """
You are tasked with summarising a long paragraph into a shorter, more concise version. 
Follow these rules strictly:

Rules:
1. Do NOT add any new information.  
2. Do NOT remove any information or meaning.  
3. Preserve all facts, relationships, and intent.  
4. The summary must be significantly more concise while remaining fully accurate.  
5. Maintain the original tone and perspective.

Use the provided examples to learn the task better.

EXAMPLE 1:
Input:
The research team conducted a six-month study to evaluate the effectiveness of the new machine learning model. They compared its performance against three existing baseline models and found that the new approach improved prediction accuracy by 15%. However, they also noted that training time increased significantly, which might limit its applicability in real-time systems."

Summary:
The six-month study found the new machine learning model improved accuracy by 15% over three baselines but required significantly longer training, limiting its real-time use.

EXAMPLE 2:
Input:
During the conference, several speakers emphasized the importance of transparent AI systems. They argued that without clear explanations of how models make decisions, user trust will remain low. Some presenters showcased tools designed to visualize model reasoning, which they claimed could help bridge the trust gap.

Summary:
Conference speakers stressed that transparent AI is essential for user trust and presented visualization tools to explain model reasoning and reduce the trust gap.

EXAMPLE 3:
Input:
The city's transportation department announced a new initiative to reduce traffic congestion by expanding bike lanes and increasing public transit frequency. Officials believe this will encourage more residents to use alternatives to driving, ultimately decreasing emissions and improving overall air quality.

Summary:
The transportation department plans to reduce congestion by expanding bike lanes and increasing transit frequency, aiming to shift residents from driving and improve emissions and air quality.

Your task:
Input:
{{context}}

Summary:
"""

INSTRUCTION_WITH_REF = """

Your task is to summarize the CONTEXT with respect to the ATOM.

Instructions: 
Follow the steps below for CONTEXT summarization:
1. The ATOM can be true, false or not verifiable according to the SUMMARY.
2. It is very possible that no relevant information about the ATOM or related to the ATOM can be found in the CONTEXT. In this case, the SUMMARY must be: "None".
3. If the CONTEXT does not provide information about the ATOM, or if the CONTEXT does not mention anything related to the ATOM, the SUMMARY must be: "None".
4. If the CONTEXT provides information about the ATOM, the SUMMARY must contain the most relevant information of the CONTEXT and be such that we can fact-check the ATOM using this SUMMARY. 
5. The SUMMARY must not use reported speech to refer to the CONTEXT, for instance the SUMMARY must NOT state: "according to the context", "this context mentions", or "this article outlines", but instead the SUMMARY must only summarize the CONTEXT.
6. If the CONTEXT provides information about the ATOM, provide the SUMMARY.
7. If the CONTEXT does not provide information about the ATOM, the SUMMARY must only provide "None". Do not mention that the context does not provide any information about the atom. Do not provide anything else.

Use the provided examples to learn the task better.

Example 1:
ATOM: Sense and Sensibility was published in the summer of 1811.

CONTEXT: + Sense and Sensibility + Sense and Sensibility is a novel by Jane \
Austen , published in 1811 . + Jane Austen + Jane Austen ( 16 December 1775 - 18 July \
1817 ) was an English novelist known primarily for her six major novels , which interpret , \
critique and comment upon the British landed gentry at the end of the 18th century .

SUMMARY:
Sense and Sensibility was published in 1811, however it is not known whether it \
has been published in summer.

Example 2:
ATOM: Filmfare is about cheese.

CONTEXT: + Filmfare + Filmfare is an English-language , tabloid-sized magazine \
about Hindi-language cinema , popularly known as Bollywood . + Bollywood + Bollywood \
is the sobriquet for India 's Hindi language film industry , based in the city of Mumbai , \
Maharashtra .

SUMMARY: 
Filmfare is about Hindi-language cinema, not about cheese.

Example 3:
ATOM: The 19th G7 summit only included Russia.

CONTEXT:
+ 19th G7 summit + The Group of Seven ( G7 ) was an unofficial forum \
which brought together the heads of the richest industrialized countries : France , Germany \
, Italy , Japan , the United Kingdom , the United States , Canada ( since 1976 ) and the \
President of the European Commission ( starting officially in 1981 ) .

SUMMARY: 
The 19th G7 summit did not only include Russia, but also the heads of the six \
other richest industrialized countries and the President of the European Commission.

Example 4:
ATOM: Quantum mechanics describes the behavior of particles at the smallest scales, where classical physics no longer applies.

CONTEXT:
The Amazon rainforest, often referred to as the "lungs of the Earth," spans over 5.5 million square kilometers across nine countries. \
It is home to millions of species, many of which are yet to be discovered. The rainforest plays a crucial role in global oxygen production \
and carbon dioxide absorption. However, it faces severe threats from deforestation, illegal mining, and climate change. Conservation efforts \
are ongoing, with governments, environmental organizations, and indigenous communities working together to protect this vital ecosystem. 

SUMMARY:
None


Example 5:
ATOM: Zeus was the creator of Nazgul.

CONTEXT:
+ Artemis + She was the Hellenic goddess of the hunt , wild animals , \
wilderness , childbirth , virginity and protector of young girls , bringing and relieving \
disease in women ; she often was depicted as a huntress carrying a bow and arrows .

SUMMARY:
None

Your task:
ATOM: {{atom_text}}

CONTEXT: {{context}}

SUMMARY:
"""

class ContextSummarizer:
    """
    Context summarization given the atom.
    """

    def __init__(
            self,
            backend: Backend,
    ):
        """
        Initialize the ContextSummarizer.

        Args:
            backend: str
                The Mellea backend to use for LLM interaction.
            with_reference: str
                The reference paragraph that the context will be summarized to.
        """
        
        # Safety checks        
        if backend is None:
            raise ValueError("Mellea backend is None. Please provide a valid Mellea backend.")

        # Initialize the extractor
        self.backend = backend
        
        # Print info
        print(f"[Summarizer] Using Mellea backend: {self.backend.model_id}")

        # Disable Mellea logging
        FancyLogger.get_logger().setLevel(FancyLogger.ERROR)

    def _get_probability(self, output: ModelOutputThunk) -> float:
        """
        Compute the average log probability of the generated tokens.

        Args:
            output: ModelOutputThunk
                The model raw output (via Mellea).

        Returns:
            float: The average log probability of the generated tokens.
        """

        assert output._meta["oai_chat_response"]["logprobs"] is not None
        logprobs = output._meta["oai_chat_response"]["logprobs"]["content"][:-1] # last token is EOS
        avg_logprob = sum(lp['logprob'] for lp in logprobs) / len(logprobs) if len(logprobs) > 0 else math.inf

        return math.exp(avg_logprob) if not math.isinf(avg_logprob) else 0.0 

    async def run_batch(self, contexts: List[str], atom_text: str = None) -> List[Dict[str, Any]]:
        """
        Summarize a list of contexts with respect to an atomic unit.
        
        Args:
            contexts: List[str]
                The list of contexts to be summarized.
            atom_text: str
                The reference atomic unit text
        Returns:
            List[Dict[str, Any]]: A list of summarized contexts.
        """

        # Initialize the instruction
        instruction = INSTRUCTION_WITH_REF if atom_text is not None else INSTRUCTION_WITHOUT_REF

        # Perform the instruction with validation
        results = []
        corutines = []
        for context in contexts:
            corutine = mfuncs.ainstruct(
                instruction,
                context=SimpleContext(),
                backend=self.backend,
                requirements=[],
                strategy=RejectionSamplingStrategy(loop_budget=LOOP_BUDGET),
                return_sampling_results=True,                    
                user_variables={"context": context, "atom_text": atom_text},
                model_options=dict(logprobs=True)
            )
            corutines.append(corutine)

        results = []
        outputs = await asyncio.gather(*(corutines[i] for i in range(len(corutines))))
        for output in outputs:
            cleaned = str(output).strip()
            results.append(
                {
                    "context": context,
                    "summary": cleaned if cleaned != "None" else "",
                    "probability": self._get_probability(output.result)
                }
            )

        return results
       



# coding=utf-8
# Copyright 2023-present the International Business Machines.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# FactReasoner pipeline

import json
import math
import os
import time
import subprocess
import uuid
import logging
import asyncio

from typing import Any, Dict

from pgmpy.factors.discrete import DiscreteFactor
from pgmpy.global_vars import logger
from pgmpy.models import MarkovNetwork
from pgmpy.readwrite import UAIWriter
from mellea.core import FancyLogger

# Local imports
from fact_reasoner.core.atomizer import Atomizer
from fact_reasoner.core.reviser import Reviser
from fact_reasoner.core.retriever import ContextRetriever
from fact_reasoner.core.summarizer import ContextSummarizer
from fact_reasoner.core.nli import NLIExtractor
from fact_reasoner.fact_graph import FactGraph
from fact_reasoner.core.base import (
    PRIOR_PROB_ATOM,
    PRIOR_PROB_CONTEXT,
    Atom,
    Context,
    Relation,
)
from fact_reasoner.core.utils import (
    build_atoms,
    build_contexts,
    build_relations,
    is_relevant_context,
    remove_duplicated_atoms,
    remove_duplicated_contexts,
)

# Set logging levels 
# pgmpy set the root logger to INFO -- changed it to WARNING
logging.getLogger("httpx").setLevel(logging.ERROR)
logger.setLevel(logging.WARNING)
class FactReasoner:
    def __init__(
            self,
            atom_extractor: Atomizer = None,
            atom_reviser: Reviser = None,
            nli_extractor: NLIExtractor = None,
            context_retriever: ContextRetriever = None,
            context_summarizer: ContextSummarizer = None,
            merlin_path: str = None,
            use_priors: bool = True,
    ):
        """
        Initialize the FactReasoner pipeline.

        Args:
            atom_extractor: Atomizer
                The service used for extracting atoms from the response.
            atom_reviser: Reviser
                The service used for decontextualizing the atoms.
            context_retriever: ContextRetriever
                The service used for retrieving external contexts.
            context_summarizer: ContextSummarizer
                The service used for summarizing contexts.
            nli_extractor: NLIExtractor
                The service used for NLI relationship extraction.
            merlin_path: str
                Path to the Merlin probabilistic reasoning engine (c++ implementation).
            use_priors: bool
                Flag indicating that atom and context priors are used in the factor definition.
        """

        # Initialize FactReasoner
        self.query = None
        self.response = None
        self.topic = None
        self.use_priors = use_priors
        self.start_time = time.perf_counter() # get the start time

        self.context_retriever = context_retriever
        self.context_summarizer = context_summarizer
        self.atom_extractor = atom_extractor
        self.atom_reviser = atom_reviser
        self.nli_extractor = nli_extractor
        self.merlin_path = merlin_path

        # Safety checks
        assert self.merlin_path is not None, f"Path to `merlin` cannot be None."

        print(f"[FactReasoner] Using merlin at: {self.merlin_path}")
        print(f"[FactReasoner] Using atom/context priors: {self.use_priors}")

        self.atoms = {}  # indexed by atom id
        self.contexts = {}  # indexed by context id
        self.relations = []

        self.num_retrieved_contexts = 0
        self.num_summarized_contexts = 0

        # The fact graph and probabilistic model (Markov Network)
        self.fact_graph = None
        self.markov_network = None

        # Ground truth labels (if any)
        self.labels_human = None

        # Disable Mellea logging
        FancyLogger.get_logger().setLevel(FancyLogger.ERROR)

    def from_fact_graph(
            self,
            fact_graph: FactGraph
    ):
        """
        Initialize FactReasoner from a FactGraph instance.

        Args:
            fact_graph: FactGraph
                A FactGraph instance.
        """

        # Create the atoms, contexts and relations
        self.atoms = {}
        self.contexts = {}
        self.relations = []

        for node_id, node in fact_graph.nodes.items():
            if node.type == "atom":
                self.atoms[node_id] = Atom(
                    id=node_id,
                    text=""
                )
            elif node.type == "context":
                self.contexts[node_id] = Context(
                    id=node_id,
                    atom=None,
                    text=""
                )

        for edge in fact_graph.edges:
            id_source = edge.source
            id_target = edge.target
            if id_source in self.atoms:
                src = self.atoms[id_source]
            elif id_source in self.contexts:
                src = self.contexts[id_source]
            if id_target in self.atoms:
                trg = self.atoms[id_target]
            elif id_target in self.contexts:
                trg = self.contexts[id_target]

            rel = Relation(
                source=src,
                target=trg,
                type=edge.type,
                probability=edge.probability,
                link=edge.link
            )

            self.relations.append(rel)

        # Create the corresponding fact graph
        self.fact_graph = fact_graph

        # Create the corresponding probabilistic model (Markov Network)
        self._build_markov_network()

    def from_dict_with_contexts(
            self,
            data: Dict[str, Any],
    ):
        """
        Initialize FactReasoner from a dict containing both atoms and contexts.

        Args:
            data: Dict[str, Any]
                The dict containing the problem instance.
        """

        self.query = data["input"]
        self.response = data["output"]
        self.topic = data.get("topic", None)

        print(f"[FactReasoner] Reading the atoms ...")
        gold_labels = []
        atom_ids = []
        self.atoms = {}
        atom2contexts = {}
        for atom_dict in data["atoms"]:
            aid = atom_dict["id"]
            text = atom_dict["text"]
            original = atom_dict["original"]
            label = atom_dict.get("label", None)
            contexts = atom_dict["contexts"]
            a = Atom(id=aid, text=text, label=label)
            a.set_original(original)
            atom_ids.append(aid)
            gold_labels.append(label)
            self.atoms[aid] = a
            atom2contexts[aid] = contexts
        
        print(f"[FactReasoner] Atoms found: {len(self.atoms.keys())}")
        for _, atom in self.atoms.items():
            print(f"[FactReasoner] {atom}")

        self.labels_human = dict(zip(atom_ids, gold_labels))
        print(f"[FactReasoner] Labels found: {self.labels_human}")

        print(f"[FactReasoner] Reading the contexts ...")
        for context_dict in data["contexts"]:
            cid = context_dict["id"]
            title = context_dict["title"]
            text = context_dict["text"]
            snippet = context_dict.get("snippet", "")
            link = context_dict.get("link", "")
            ctxt = Context(
                id=cid, 
                atom=None, 
                text=text, 
                title=title, 
                snippet=snippet, 
                link=link
            )
            self.contexts[cid] = ctxt

        print(f"[FactReasoner] Contexts found: {len(self.contexts)}")
        for aid, atom in self.atoms.items():
            ctxts = []
            for c in atom2contexts[aid]:
                ctxts.append(self.contexts[c])
                self.contexts[c].set_atom(atom)
            atom.add_contexts(ctxts)
        
        print(f"[FactReasoner] Pipeline initialized with {len(self.atoms)} atoms and {len(self.contexts)} contexts.")

    def build(
            self,
            query: str = None,
            response: str = None,
            topic: str = None,
            has_atoms: bool = False,
            has_contexts: bool = False,
            revise_atoms: bool = False,
            remove_duplicates: bool = False,
            summarize_contexts: bool = True,
            contexts_per_atom_only: bool = False,
            rel_atom_context: bool = True,
            rel_context_context: bool = False,
            use_fast_retriever: bool = True
    ):
        """
        Build the atoms and contexts using the retrieval service.

        Args:
            query: str
                The input user query.
            response: str
                The LLM generated response to the input query.
            topic: str
                The topic of the input query/response.
            has_atoms: bool
                Flag indicating if the atoms were previously initialized.
            has_contexts: bool
                Flag indicating if the contexts were previously initialized.
            revise_atoms: bool
                Flag indicating that the atoms will be revised (decontextualized).
            remove_duplicates: bool
                Flag indicating if duplicated contexts are to be removed.
            summarize_contexts: bool
                Flag indicating if contexts are to be summarized.
            contexts_per_atom_only: bool
                Flag indicating that only the contexts retrieved per atom will be used.
            rel_atom_context: bool (default is True)
                Flag indicating the presence of atom-to-context relationships.
            rel_context_context: bool (default is False)
                Flag indicating the presence of context-to-context relationships.
        """

        # Initialize the reasoner
        if query is not None: 
            self.query = query
        if response is not None:
            self.response = response
        if topic is not None:
            self.topic = topic

        self.fact_graph = None
        self.markov_network = None
        self.revise_atoms = revise_atoms
        self.summarize_contexts = summarize_contexts # default is False

        # Safety checks
        assert self.atom_extractor is not None, \
            f"The atom extractor must be created."
        assert self.atom_reviser is not None, \
            f"The atom reviser must be created."
        assert self.nli_extractor is not None, \
            f"The NLI extractor must be created."

        print(f"[FactReasoner] Building the pipeline ...")
        
        # Build the atoms
        if has_atoms == False:
            self.atoms = build_atoms(
                response=self.response,
                atom_extractor=self.atom_extractor
            )
            self.revise_atoms = True # revise the atoms if newly created
            print(f"[FactReasoner] Extracted {len(self.atoms)} atoms.")
            for aid in self.atoms.keys():
                print(f"[FactReasoner] {self.atoms[aid]}")

        # Safety checks
        assert (len(self.atoms) > 0), \
            f"The atoms must be initialized before running the pipeline."

        # Revise the atoms
        if self.revise_atoms:
            print(f"[FactReasoner] Revising the atoms ...")
            assert self.response is not None, f"The atom reviser requires a response."
            atom_ids = [aid for aid in sorted(self.atoms.keys())]
            old_atoms = [self.atoms[aid].get_text() for aid in atom_ids]
            result = self.atom_reviser.run(old_atoms, self.response)
            for i, aid in enumerate(atom_ids):
                elem = result[i]
                self.atoms[aid].set_text(elem["revised_unit"])
                print(f"[FactReasoner] {self.atoms[aid]}")

        # Remove duplicated atoms (if any)
        self.atoms = remove_duplicated_atoms(self.atoms)
        print(f"[FactReasoner] Created {len(self.atoms)} unique atoms.")

        # Build the contexts (per atom)
        if has_contexts == False: # check if contexts already in file
            self.contexts = build_contexts(
                atoms=self.atoms,
                query=self.query,
                retriever=self.context_retriever,
                use_fast_retriever=use_fast_retriever
            )

        # For tracking purposes
        self.num_retrieved_contexts = len(self.contexts.keys()) 
        print(f"[FactReasoner] Retrieved {self.num_retrieved_contexts} contexts.")

        # Safety checks
        assert (len(self.contexts.keys()) > 0 or not has_contexts), \
            f"Contexts must be initialized if `has_contexts` is True!"
        
        # Remove duplicated contexts
        if remove_duplicates:
            self.contexts, self.atoms = remove_duplicated_contexts(self.contexts, self.atoms)
            print(f"[FactReasoner] Created {len(self.contexts.keys())} unique contexts.")

        # Summarize the retrieved contexts (if any)       
        if self.summarize_contexts:
            print(f"[FactReasoner] Summarizing the contexts ...")
            
            # Summarize contexts for atoms
            for atom_id, atom in self.atoms.items():
                if len(atom.contexts.keys()) > 0:
                    contexts_ids, contexts =  zip(*atom.contexts.items()) 
                    results = asyncio.run(self.context_summarizer.run_batch([context.get_text() for context in contexts], atom.text)) 

                    # Safety checks
                    assert len(results) == len(contexts), f"The number of summaries must be equal to the number of contexts."
                    
                    # Set the new syntheric summaries
                    for context_id, result in zip(contexts_ids, results):
                        is_relevant = is_relevant_context(result["summary"])
                        if result["summary"] != "" and is_relevant:
                            self.contexts[context_id].set_synthetic_summary(result["summary"])
                            # update prior probability of context based on the confidence estimation of the summary
                            self.contexts[context_id].set_probability(result["probability"] * self.contexts[context_id].get_probability())
                        else:
                            # we remove the context because it is not related to the atom
                            del self.contexts[context_id]
                            del self.atoms[atom_id].contexts[context_id]
                    print(f"[FactReasoner] Created {len(results)} summarized contexts for atom {atom_id}.")

            # Summarize contexts for question
            c_qs = {c_id: context for c_id, context in self.contexts.items() if c_id.startswith("c_q")} 
            if len(c_qs.keys()) > 0:
                contexts_ids, contexts =  zip(*c_qs.items()) 
                results = asyncio.run(self.context_summarizer.run_batch([context.get_text() for context in contexts], self.query))

                assert len(results) == len(contexts), f"The number of summaries must be equal to the number of contexts."
                
                for context_id, result in zip(contexts_ids, results):
                    is_relevant = is_relevant_context(result["summary"])
                    if result["summary"] != "" and is_relevant:
                        self.contexts[context_id].set_synthetic_summary(result["summary"])
                        # update prior probability of context based on the confidence estimation of the summary
                        self.contexts[context_id].set_probability(result["probability"] * self.contexts[context_id].get_probability())
                    else:
                        # we remove the context because it is not related to the atom
                        del self.contexts[context_id]
                print(f"[FactReasoner] Created {len(results)} summarized contexts for the question.")

            # For tracking purposes
            self.num_summarized_contexts = len(self.contexts.keys()) 
            print(f"[FactReasoner] Created {self.num_summarized_contexts} summarized contexts.")

            # Remove duplicated contexts that have the same summary (if any)
            if remove_duplicates:
                self.contexts, self.atoms = remove_duplicated_contexts(self.contexts, self.atoms, check_summary=True)
                print(f"[FactReasoner] Created {len(self.contexts.keys())} unique summarized contexts.")


        # Build the NLI relationships
        self.relations = build_relations(
            atoms=self.atoms,
            contexts=self.contexts,
            rel_atom_context=rel_atom_context,
            rel_context_context=rel_context_context,
            contexts_per_atom_only=contexts_per_atom_only,
            nli_extractor=self.nli_extractor,
            use_summarized_contexts=self.summarize_contexts,
        )

        # Build the fact graph and Markov network
        print(f"[FactReasoner] Building the graphical model ...")
        self._build_fact_graph()
        self._build_markov_network()

        print(f"[FactReasoner] Pipeline instance created.")

    def to_json(self, json_file_path: str = None) -> Dict[str, Any]:
        """
        Save the FactReasoner instance to a JSON file.

        Args:
            json_file: str
                The path to the output JSON file.
        """

        data = {}
        data["input"] = self.query
        data["output"] = self.response
        data["topic"] = self.topic
        data["atoms"] = []
        data["contexts"] = []

        # Write the atoms
        for aid, atom in self.atoms.items():
            atom_data = dict(
                id=aid, text=atom.get_text(), contexts=list(atom.get_contexts().keys())
            )
            if atom.get_label() is not None:
                atom_data["label"] = atom.get_label()
            data["atoms"].append(atom_data)

        # Write the contexts
        data["contexts"] = [context.to_json() for context in self.contexts.values()]

        # Write to a JSON file (if any)
        if json_file_path:
            with open(json_file_path, "w") as f:
                json.dump(data, f, indent=4)
            f.close()
            print(f"[FactReasoner] Pipeline instance written to: {json_file_path}")

        return data

    def _build_fact_graph(self):
        """
        Create the fact graph representation from atoms, contexts and relations.
        """

        self.fact_graph = FactGraph(
            atoms=list(self.atoms.values()),
            contexts=list(self.contexts.values()),
            relations=self.relations
        )

    def _build_markov_network(self):
        """
        Create the Markov Network corresponding to the FactGraph.

        Return:
            A MarkovNetwork encoding of the problem.
        """

        assert self.fact_graph is not None, f"The FactGraph must be built."

        # Create an empty Markov Network
        self.markov_network = MarkovNetwork()

        # Create the variables corresponding to the nodes in the fact graph
        print(f"[Building the Markov network...]")
        for node in self.fact_graph.get_nodes():
            x = node.id
            self.markov_network.add_node(x)
            if node.type == "context":
                prob = node.probability  #PRIOR_PROB_CONTEXT
                factor = DiscreteFactor(
                    variables=[x],
                    cardinality=[2],
                    values=[1.0 - prob, prob]
                )
                self.markov_network.add_factors(factor)
                print(f"Adding context variable {x} with discrete factor (prior)")
            elif node.type == "atom":
                prob = node.probability  # PRIOR_PROB_ATOM
                factor = DiscreteFactor(
                    variables=[x],
                    cardinality=[2],
                    values=[1.0 - prob, prob]
                )
                self.markov_network.add_factors(factor)
                print(f"Adding atom variable {x} with discrete factor (prior)")
            else:
                raise ValueError(f"Unknown node type: {node.type}")

        # Create the factors corresponding to the edges in the fact graph
        for edge in self.fact_graph.get_edges():
            x, y = edge.source, edge.target
            self.markov_network.add_edge(x, y)
            if edge.type == "entailment":  # add factor X -> Y
                prob = edge.probability
                if self.use_priors:
                    if edge.link == "context_atom":
                        values = [1.0 - PRIOR_PROB_ATOM, PRIOR_PROB_ATOM, 1.0 - prob, prob]
                    elif edge.link == "context_context":
                        values = [1.0 - PRIOR_PROB_CONTEXT, PRIOR_PROB_CONTEXT, 1.0 - prob, prob]
                    elif edge.link == "atom_atom":
                        values = [1.0 - PRIOR_PROB_ATOM, PRIOR_PROB_ATOM, 1.0 - prob, prob]
                    else:
                        raise ValueError(f"Unknown link type: {edge.link}")
                else:
                    values = [prob, prob, 1.0 - prob, prob]

                # Create the factor    
                factor = DiscreteFactor(
                    variables=[x, y],
                    cardinality=[2, 2],
                    values=values  #[prob, prob, 1.0 - prob, prob]
                )
                self.markov_network.add_factors(factor)
                print(f"Adding edge {x} - {y} with discrete factor (entailment)")
            elif edge.type == "contradiction":  # add factor X -> !Y
                prob = edge.probability
                if self.use_priors:
                    if edge.link == "context_atom":
                        values = [1.0 - PRIOR_PROB_ATOM, PRIOR_PROB_ATOM, prob, 1.0 - prob]
                    elif edge.link == "context_context":
                        values = [1.0 - PRIOR_PROB_CONTEXT, PRIOR_PROB_CONTEXT, prob, 1.0 - prob]
                    elif edge.link == "atom_atom":
                        values = [1.0 - PRIOR_PROB_ATOM, PRIOR_PROB_ATOM, prob, 1.0 - prob]
                    else:
                        raise ValueError(f"Unknown link type: {edge.link}")
                else:
                    values = [prob, prob, prob, 1.0 - prob]

                factor = DiscreteFactor(
                    variables=[x, y],
                    cardinality=[2, 2],
                    values=values  #[prob, prob, prob, 1.0 - prob]
                )
                self.markov_network.add_factors(factor)
                print(f"Adding edge {x} - {y} with discrete factor (contradiction)")
            elif edge.type == "equivalence":
                prob = edge.probability
                factor = DiscreteFactor(
                    variables=[x, y],
                    cardinality=[2, 2],
                    values=[prob, 1.0 - prob, 1.0 - prob, prob]
                )
                self.markov_network.add_factors(factor)
                print(f"Adding edge {x} - {y} with discrete factor (equivalence)")

        # Output the content of the network
        print("[Markov network created.]")

    def run_merlin(self):
        """
        Run inference with merlin (executable)
        """

        # Prepare the query variables (i.e., atoms)
        query_variables = [var for var in sorted(self.atoms.keys())]

        # Dump the markov network to a temporary file
        net_id = str(uuid.uuid1())
        input_filename = f"markov_network_{net_id}.uai"
        writer = UAIWriter(self.markov_network)
        writer.write_uai(input_filename)

        # Get the variable name to index mapping {0: ('a0', '2'), 1: ('a1', '2')}
        vars_mapping = {}
        variables = sorted(writer.domain.items(), key=lambda x: (x[1], x[0]))
        for i, var in enumerate(variables):
            vars_mapping[i] = var[0]

        # Run merlin as a subprocess and collect the results
        exefile = self.merlin_path
        output_format = "json"
        output_file = f"output_{net_id}"
        algorithm = "wmb"
        task = "MAR"

        args = [
            exefile,
            "--input-file", input_filename,
            "--task", task,
            "--ibound", "6",
            "--algorithm", algorithm,
            "--output-format", output_format,
            "--output-file", output_file
        ]

        proc = subprocess.run(args)

        print(f"[Merlin] return code: {proc.returncode}")
        output_filename = f"{output_file}.{task}.{output_format}"
        with open(output_filename) as f:
            results = json.load(f)

        marginals = []
        all_marginals = []
        for marginal in results["marginals"]:
            var_index = marginal["variable"]
            var_name = vars_mapping[var_index]
            all_marginals.append(dict(variable=var_name, probabilities=marginal["probabilities"]))
            if var_name in query_variables:
                probs = marginal["probabilities"]
                marginals.append({"variable": var_name, "probabilities": probs})

        # Cleanup -- delete input_filename and output_filename
        if os.path.exists(input_filename): os.remove(input_filename)
        if os.path.exists(output_filename): os.remove(output_filename)

        print(f"[Merlin] All Marginals:\n{all_marginals}")
        return marginals

    def score(self) -> Dict[str, Any]:
        """
        Compute the factuality score taking into consideration the contexts 
        retrieved for each of the atom in the answer.

        Factuality score = # atoms(true) / # atoms

        Intuitively, a score of 100% means that all atoms in the answer are
        factually correct. If none of them are correct, then the score is 0%. If
        only half of the atoms are correct, then the score is 50%.

        Returns:
            Dict[str, Any]: The results dictionary containing the marginals, factuality score i.e., a real value in [0, 1]
        """

        # Safety checks
        if len(self.atoms.keys()) == 0:
            print("WARNING: no atoms have been identified!")
        if len(self.contexts.keys()) == 0:
            print("WARNING: no contexts have been retrieved!")
        if len(self.relations) == 0:
            print("WARNING: no relationships have been identified!")

        assert self.fact_graph is not None
        assert self.markov_network is not None

        marginals = self.run_merlin()

        # Prepare the results
        num_true_atoms = 0
        num_uniform_atoms = 0
        avg_prob = 0.0
        avg_logprob = 0.0
        entropy = 0.0
        norm_entropy = 0.0
        avg_norm_entropy = 0.0
        labels = {}
        probabilities = {}
        fscore_per_atom = []
        elapsed_time = time.perf_counter() - self.start_time # total elapsed time
        for marginal in marginals:
            var = marginal["variable"]
            probs = marginal["probabilities"]

            print(f"[FactReasoner] ({var}): Probability for {var}=0 is: {probs[0]}")
            print(f"[FactReasoner] ({var}): Probability for {var}=1 is: {probs[1]}")

            # Check if atom is true or not
            probabilities[var] = probs[1] # probability of true
            if probs[1] > probs[0]:
                num_true_atoms += 1
                labels[var] = "S"
            else:
                labels[var] = "NS"

            fscore_per_atom.append({var: {"score": probs[1], "support": labels[var]}})
            probval = probs[1]
            if probval < 1e-6:
                probval = 1e-6
            elif probval >= 1.0:
                probval = 0.999999
            elif probval == 0.5:
                num_uniform_atoms += 1
            avg_logprob += math.log(probval)
            avg_prob += probval
            entropy += -probval * math.log(probval)
            norm_entropy += -(probval * math.log(probval) + (1.0 - probval) * math.log(1.0 - probval)) / math.log(2.0)

        # For now, return a dict with the posterior marginals of the atoms
        avg_logprob /= len(self.atoms)
        avg_prob /= len(self.atoms)
        avg_entropy = entropy / len(self.atoms)
        avg_norm_entropy = norm_entropy / len(self.atoms)
        fscore = num_true_atoms / len(self.atoms)

        # Precision, R@K and F1@K
        fscore = float(num_true_atoms)/float(len(self.atoms))
        K = int(len(self.atoms) / 2) # K is assumed to be half
        recall_k = min(float(num_true_atoms/K), 1.0)
        try:
            f1k = 2 * fscore * recall_k / (fscore + recall_k)
        except Exception as _:
            f1k = 0.0

        # Elapsed time
        elapsed_time = time.perf_counter() - self.start_time # elapsed time

        results = {}
        results["factuality_score_per_atom"] = fscore_per_atom
        results["factuality_score"] = fscore
        results["recall_k"] = recall_k
        results["f1_k"] = f1k
        results["num_atoms"] = len(self.atoms)
        results["num_contexts"] = len(self.contexts)
        results["num_true_atoms"] = num_true_atoms
        results["num_false_atoms"] = len(self.atoms) - num_true_atoms
        results["num_uniform_atoms"] = num_uniform_atoms
        results["entropy"] = entropy
        results["norm_entropy"] = norm_entropy
        results["avg_entropy"] = avg_entropy
        results["avg_norm_entropy"] = avg_norm_entropy
        results["avg_prob"] = avg_prob
        results["avg_logprob"] = avg_logprob # math.exp(avg_logprob)
        results["avg_explogprob"] = math.exp(avg_logprob)
        results["marginals"] = marginals
        results["predictions"] = labels
        print(f"[FactReasoner] Predictions: {labels}")

        # Remove duplicate atoms in self.labels_human
        if self.labels_human is not None:
            self.labels_human = {k: v for i, (k, v) in enumerate(self.labels_human.items()) if k in self.atoms.keys() and k not in list(self.labels_human.keys())[:i]}

        # Check for ground truth annotations
        if self.labels_human is not None:
            true_atoms = 0
            false_atoms = 0
            avg_brier = 0.0
            num_true_positive = 0
            num_true_negative = 0
            num_false_positive = 0
            num_false_negative = 0
            for aid, l in self.labels_human.items():

                if l == "S":
                    avg_brier += (probabilities[aid] - 1.0) * (probabilities[aid] - 1.0)
                    true_atoms += 1
                    if labels[aid] == "S":
                        num_true_positive += 1
                    else:
                        num_false_negative += 1
                else:
                    avg_brier += (probabilities[aid] - 0.0) * (probabilities[aid] - 0.0)
                    false_atoms += 1
                    if labels[aid] == "NS":
                        num_true_negative += 1
                    else:
                        num_false_positive += 1
            fscore_gold = true_atoms / len(self.labels_human.keys())
            avg_brier /= len(self.atoms)
            print(f"[FactReasoner] Gold labels: {self.labels_human}")
            print(f"[FactReasoner] Gold fscore: {fscore_gold} ({true_atoms}/{len(self.labels_human.keys())})")
            results["gold_factuality_score"] = fscore_gold
            results["gold_true_atoms"] = true_atoms
            results["true_positive"] = num_true_positive
            results["true_negative"] = num_true_negative
            results["false_positive"] = num_false_positive
            results["false_negative"] = num_false_negative
            results["references"] = self.labels_human
            results["avg_brier"] = avg_brier

        results["topic"] = self.topic
        results["query"] = self.query
        results["response"] = self.response
        results["elapsed_time"] = elapsed_time
        print(f"[FactReasoner] Elapsed time: {elapsed_time:.4f} seconds.")

        return results, marginals




"""Factuality Assessment Baselines."""

from .factscore import FactScore
from .factverify import FactVerify
from .veriscore import VeriScore

__all__ = [
    "FactScore",
    "FactVerify",
    "VeriScore",
]


# coding=utf-8
# Copyright 2023-present the International Business Machines.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# FactCorrector pipeline

import json
import math
import os
import time
import subprocess
import uuid
import logging

from typing import Any, Dict

from pgmpy.factors.discrete import DiscreteFactor
from pgmpy.global_vars import logger
from pgmpy.models import MarkovNetwork
from pgmpy.readwrite import UAIWriter
from mellea.backends import ModelOption
from mellea.core import FancyLogger

class FactCorrector:
    def __init__(self):
        pass


# coding=utf-8
# Copyright 2023-present the International Business Machines.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re
import chromadb
import requests
import torch
import asyncio
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Dict, List, Optional

from tqdm import tqdm

from bs4 import BeautifulSoup
from chromadb.utils import embedding_functions
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Any, Dict, List, Optional
from io import BytesIO
from PyPDF2 import PdfReader
from itertools import islice

logger = logging.getLogger(__name__)
logger.setLevel(logging.CRITICAL)

from langchain_text_splitters import RecursiveCharacterTextSplitter
from langchain_community.retrievers import WikipediaRetriever
from langchain_community.vectorstores import InMemoryVectorStore
from langchain_core.documents import Document
from langchain_huggingface import HuggingFaceEmbeddings

# Local imports
from fact_reasoner.core.summarizer import ContextSummarizer
from fact_reasoner.core.query_builder import QueryBuilder
from fact_reasoner.core.base import Atom, Context
from fact_reasoner.search_api import SearchAPI

DEFAULT_COLLECTION_NAME = "lit_agent_demo"
DEFAULT_DB_PATH = "/tmp/nasa_contrib/accelerated-discovery/chroma_db"

EMBEDDING_MODEL = "all-MiniLM-L6-v2"
NEWLINES_RE = re.compile(r"\n{2,}")  # two or more "\n" characters

CHARACTER_SPLITTER = RecursiveCharacterTextSplitter(
    separators=["\n\n", "\n", ". ", " ", ""],
    #keep_separator=False,
    chunk_size=1000,
    chunk_overlap=0
)

# Regex patterns to remove common inline citation forms
CITATION_PATTERNS = [
    r"\[\d+\]",                       # [1], [23]
    r"\(\d+\)",                       # (1), (23)
    r"\[\s*citation\s+needed\s*\]",   # [citation needed]
    r"\(\s*citation\s+needed\s*\)",   # (citation needed)
    r"\[\s*[A-Za-z]+(?:\s+[A-Za-z]+)*\s+\d{4}\s*\]",  # [Smith 2020], [Doe et al. 2019]
    r"\^\d+",                         # ^1 (superscript-like)
]

def _clean_text(text: str) -> str:
    """Apply citation removal and whitespace normalization (single line)."""
    for pattern in CITATION_PATTERNS:
        text = re.sub(pattern, " ", text, flags=re.IGNORECASE)
    # Collapse all whitespace (including newlines/tabs) to a single space
    text = re.sub(r"\s+", " ", text).strip()
    return text

def _extract_html_paragraphs(soup: BeautifulSoup) -> str:
    """
    Extract only paragraph text (<p> tags) from HTML, removing common citation elements
    and returning a single-line string.
    """
    # Remove non-content tags globally (helps reduce clutter if nested inside <p>)
    for tag_name in ("script", "style", "noscript"):
        for tag in soup.find_all(tag_name):
            tag.decompose()

    # Remove typical citation nodes seen on many sites (e.g., Wikipedia)
    for el in soup.find_all(["sup", "span"], class_=lambda c: c and "reference" in c):
        el.decompose()
    for el in soup.select(".citation, .footnote, .ref, .references"):
        el.decompose()

    # Collect visible text only from <p> tags
    paragraphs = []
    for p in soup.find_all("p"):
        # Defensive: remove nested citation-like elements inside the paragraph
        for el in p.find_all(["sup", "span"], class_=lambda c: c and "reference" in c):
            el.decompose()
        for el in p.select(".citation, .footnote, .ref, .references"):
            el.decompose()

        para_text = p.get_text(separator=" ", strip=True)
        if para_text:
            paragraphs.append(para_text)

    # Join paragraphs into one line (no newlines), removing citations and normalizing whitespace
    text = " ".join(_clean_text(p) for p in paragraphs)
    return _clean_text(text)

def _extract_pdf_paragraphs(pdf_bytes: bytes, max_pages: int = 1) -> str:
    """
    Extract text from a PDF and return a single-line string.

    - If max_pages is provided, only the first max_pages pages are parsed.
    - Uses a simple heuristic to detect paragraphs (splitting on blank lines or multiple newlines).
    """
    reader = PdfReader(BytesIO(pdf_bytes))

    # Decide how many pages to read
    total_pages = len(reader.pages)
    pages_iter = (reader.pages[i] for i in range(total_pages))
    if max_pages is not None and max_pages > 0:
        pages_iter = islice(pages_iter, max_pages)

    pages_text = []
    for page in pages_iter:
        page_text = page.extract_text() or ""
        pages_text.append(page_text)

    full_text = "\n".join(pages_text)

    # Heuristic paragraph splitting: blank lines or >=2 newlines often separate paragraphs.
    raw_paragraphs = re.split(r"\n\s*\n|\r\n\s*\r\n|(?:\n{2,})", full_text)

    # Fallback: if no obvious paragraph boundaries, just use the whole text
    if not raw_paragraphs or len(raw_paragraphs) == 1:
        cleaned = _clean_text(full_text)
        return cleaned

    cleaned_paragraphs = [ _clean_text(p) for p in raw_paragraphs if p and p.strip() ]
    return _clean_text(" ".join(cleaned_paragraphs))

def extract_text_from_url(url: str, timeout: int = 10, max_pages: int = 1) -> str:
    """
    Extract text from a given URL.

    - HTML: returns ONLY the text inside <p> tags; removes citations; single-line output.
    - PDF: uses a simple paragraph heuristic; single-line output.
    - Avoids resource leaks by closing the response.

    Args:
        url (str): The web link to extract text from.
        timeout (int): Request timeout in seconds (default: 20).

    Returns:
        str: Single-line cleaned text (paragraphs-only for HTML).
    """
    headers = {
        "User-Agent": (
            "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 "
            "(KHTML, like Gecko) Chrome/122.0 Safari/537.36"
        )
    }

    response = None
    try:
        response = requests.get(url, stream=True, timeout=timeout, headers=headers)
        response.raise_for_status()
        content_type = (response.headers.get("Content-Type") or "").lower()

        # PDF branch
        if "pdf" in content_type or url.lower().endswith(".pdf"):
            # Use response.content to avoid manual streaming complexity; small PDFs fit fine in memory.
            return _extract_pdf_paragraphs(response.content, max_pages=max_pages)

        # HTML branch
        soup = BeautifulSoup(response.content, "html.parser")
        return _extract_html_paragraphs(soup)

    except Exception as e:
        logger.error(f"Error extracting text: {e}")
        return ""

    finally:
        # Ensure response is closed to avoid memory/resource leaks
        try:
            if response is not None:
                response.close()
        except Exception:
            pass

def fetch_text_from_link(link: str, max_size: int = None) -> str:
    logger.info(f"Fetching text from link: {link}")
    url_text = extract_text_from_url(url=link)
    if max_size is not None and len(url_text) > max_size:
        url_text = url_text[:max_size]
    return url_text    

def get_title(text: str) -> str:
    """
    Get the title of the retrived document. By definition, the first line in the
    document is the title (we embedded them like that).
    """
    return text[:text.find("\n")]

def make_uniform(text: str) -> str:
    """
    Return a uniform representation of the text using the langchain textsplitter
    and tokensplitter tools.
    """

    character_split_texts = CHARACTER_SPLITTER.split_text(text)
    return " ".join(character_split_texts)

class ChromaReader:
    def __init__(
        self,
        collection_name: str,
        persist_directory: str,
        embedding_model: str,
        collection_metadata: dict = None,
    ):
        """
        Initialize the ChromaDB.

        Args:
            collection_name: str
                The collection name in the vector database.
            persist_directory: str
                The directory used for persisting the vector database.
            embedding_model: str
                The embedding model.
            collection_metadata: dict
                A dict containing the collection metadata.
        """
        
        self.device = "cpu"
        if torch.cuda.is_available():
            self.device = "cuda"

        self.client = chromadb.PersistentClient(path=persist_directory)
        self.persist_directory = persist_directory
        self.collection_name = collection_name
        self.embedding_model = embedding_model
        self.embedding_function = embedding_functions.SentenceTransformerEmbeddingFunction(
            model_name=self.embedding_model,
            device=self.device,
        )
        self.collection = self.client.get_or_create_collection(
            name=self.collection_name,
            embedding_function=self.embedding_function,
            metadata=collection_metadata,
        )

    def is_empty(self):
        return self.collection.count() == 0

    def query(self, query_texts: str, n_results: int = 5):
        """
        Returns the closests vector to the question vector
        
        Args:
            query_texts: str
                The user query text.
            n_results: int
                The number of results to generate.

        Returns
            The closest result to the given question.
        """
        return self.collection.query(query_texts=query_texts, n_results=n_results)


def is_content_valid(link: str, page_text: str) -> bool:
    """
    Checks if the scraped page text is likely to be valid content
    and not a restriction notice, boilerplate, or garbled/binary data.
    """
    # Initial check for empty or invalid input
    if not page_text or not isinstance(page_text, str):
        return False

    # --- Check 1: Look for specific "red flag" phrases ---
    restriction_phrases = [
        "please contact our support team",
        "cookies are used by this site",
        "copyright ©",
        "all rights are reserved",
        "ai training, and similar technologies",
        "enable javascript",
        "must have javascript enabled",
        "to continue, please verify you are a human",
        "access denied",
        "403 forbidden",
        "errors.edgesuite",
        "you do not have access to"
    ]
    
    text_lower = page_text.lower()
    for phrase in restriction_phrases:
        if phrase in text_lower:
            logger.warning(f"Redundant content detected for {link} due to restriction phrase: '{phrase}'")
            return False

    if len(page_text) > 50:
        # TODO: add more filters
        replacement_char_count = page_text.count("�")
        try:
            ratio = replacement_char_count / len(page_text)
        except ZeroDivisionError:
            ratio = 0
        if ratio > 0.10:
            logger.warning(f"Redundant content detected for {link} due to high ratio of replacement characters: {ratio:.2%}")
            return False

    # we consider the content as valid
    return True


class Retriever:
    """
    The Retriever component. We implement several versions of this component
    using a remote chromadb store (API exists), a local chromadb store, langchain
    based wikipedia retriever, and possibly others.
    """
    
    def __init__(
            self,
            service_type: str = "chromadb",
            collection_name: str = "wikipedia_en",
            persist_dir: str = "/tmp/wiki_db",
            top_k: int = 1,
            cache_dir: Optional[str] = None,
            fetch_text: bool = False,
            use_in_memory_vectorstore: bool = False,
            query_builder: QueryBuilder = None,
            num_workers: int = 4
    ):
        """
        Initialize the context retriever component.

        Args:
            service_type: str
                The type of the context retriever (chromadb, wikipedia, google)
            collection_name: str
                Name of the collection of documents stored in the vectorstore
            persist_directory: str
                The dir name where the vectorstore is persisted
            top_k: int
                The top k most relevant contexts.
            cache_dir: str
                Path to the folder containing the cache (db, json).
            fetch_text: bool
                Flag to retrieve content from a link.
            use_in_memory_vectorstore: bool
                Flag to use an in memory vectorstore over chunks of retrieved texts when using Google retriever.
                Use in cases where the search results contain long documents so that they can be broken up
                into smaller chunks, which will be retrieved using the query text. When disbaled (default)
                the input will be truncated to a `max_size`.
            query_builder: QueryBuilder
                An instance of QueryBuilder to generate search queries.
            num_workers: int
                Number of threads for parallel link fetching (default: 4).
        """

        self.top_k = top_k
        self.num_workers = num_workers
        self.service_type = service_type
        self.cache_dir = cache_dir
        self.persist_dir = persist_dir
        self.fetch_text = fetch_text
        self.use_in_memory_vectorstore = use_in_memory_vectorstore
        self.query_builder = query_builder
        self.collection_name = collection_name

        self.chromadb_retriever = None
        self.langchain_retriever = None
        self.google_retriever = None
        self.in_memory_vectorstore = None

        assert self.service_type in ["chromadb", "wikipedia", "google"]

        if self.service_type == "chromadb":
            self.chromadb_retriever = ChromaReader(
                collection_name=self.collection_name, 
                persist_directory=self.persist_dir, 
                embedding_model=EMBEDDING_MODEL, 
                collection_metadata={"hnsw:space": "cosine"}
            )
        elif self.service_type == "wikipedia":
            # Create the Wikipedia retriever. Note that page content is capped
            # at 4000 chars. The metadata has a `title` and a `summary` of the page.
            self.langchain_retriever = WikipediaRetriever(lang="en", top_k_results=top_k)
        elif self.service_type == "google":
            self.google_retriever = SearchAPI(cache_dir=self.cache_dir)
            if self.use_in_memory_vectorstore:
                self.in_memory_vectorstore = InMemoryVectorStore(
                    HuggingFaceEmbeddings(model_name=EMBEDDING_MODEL)
                )
        else:
            raise ValueError(f"Unknown retriever service: {self.service_type}")
        
    def set_query_builder(self, query_builder: QueryBuilder = None):
        self.query_builder = query_builder
    
    def query(
            self, 
            text: str,
            max_size: int = 4000
    ) -> List[Dict[str, Any]]:
        """
        Retrieve a number of contexts relevant to the input text.

        Args:
            text: str
                The input query text.

        Returns:
            List[Dict[str, Any]]:
                The list of retrieved contexts for the input reference. A context
                is a dict with 4 keys: title, text, snippet and link.
        """

        results = []
        if self.service_type == "chromadb":
            logger.info(f"Retrieving {self.top_k} relevant documents for query: {text} (service: {self.service_type})")

            relevant_chunks = self.chromadb_retriever.query(
                query_texts=[text],
                n_results=self.top_k,
            )

            docs = relevant_chunks.get("documents", [[]])[0]
            metadatas = relevant_chunks.get("metadatas", [[]])[0]
            
            passages = []
            for doc_content, metadata in zip(docs, metadatas):
                passage = {
                    "title": metadata.get("title", "No Title Provided"),
                    "text": make_uniform(doc_content),
                    "snippet": "",
                    "link": metadata.get("source", "")
                }
                passages.append(passage)

            results.extend(passages)
        elif self.service_type == "wikipedia":
            logger.info(f"Retrieving {self.top_k} relevant documents for query: {text} (service: {self.service_type})")

            passages = []                

            # Get most relevant docs to the query
            rel_docs = self.langchain_retriever.invoke(text)
            for doc in rel_docs:
                title = doc.metadata["title"]
                summary = doc.metadata["summary"]
                link = doc.metadata["source"]
                doc_content = make_uniform(doc.page_content)
                passages.append(dict(title=title, text=doc_content, snippet=summary, link=link))
            
            # Extract the top_k passages
            n = min(self.top_k, len(passages))
            for i in range(n):
                results.append(passages[i]) # a passage is a dict with title and text as keys

        elif self.service_type == "google":
            logger.info(f"Retrieving {self.top_k} search results for: {text} (service: {self.service_type})")

            if not text:
                return results # empty list

            # Generate the query text if there is a query builder
            if self.query_builder is not None:
                query_text = self.query_builder.run(text)
            else:
                query_text = text
            
            # Truncate the text if too long (for Google)
            query_text = query_text if len(query_text) < 2048 else query_text[:2048]
            logger.info(f"Using query text: {query_text}")
            passages = []

            # Get the search results
            search_results = self.google_retriever.get_snippets([query_text])

            n = len(search_results[query_text])

            # If no hits then relax query by removing specific '"' (if any)
            if n == 0: # no hits
                query_text = query_text.replace('"', '') # relax query text
                search_results = self.google_retriever.get_snippets([query_text])

            search_hits = search_results[query_text]
            n = len(search_hits)

            # --- Parallel fetch all links ---
            if self.fetch_text:
                max_sz = None if self.use_in_memory_vectorstore else max_size
                with ThreadPoolExecutor(max_workers=self.num_workers) as executor:
                    fetch_futures = {
                        executor.submit(fetch_text_from_link, hit["link"], max_sz): i
                        for i, hit in enumerate(search_hits)
                    }
                    fetched_texts = [""] * n
                    for future in as_completed(fetch_futures):
                        idx = fetch_futures[future]
                        try:
                            fetched_texts[idx] = future.result()
                        except Exception:
                            fetched_texts[idx] = ""
            else:
                fetched_texts = [""] * n

            # --- Select top_k valid results ---
            passages = []
            index_available = []
            count_content = 0

            for i, hit in enumerate(search_hits):
                if count_content >= self.top_k:
                    break

                title, snippet, link = hit["title"], hit["snippet"], hit["link"]
                raw_page_text = fetched_texts[i]

                if self.fetch_text:
                    if not is_content_valid(link, raw_page_text):
                        raw_page_text = ""
                    doc_content = make_uniform(raw_page_text) if raw_page_text else ""

                    if not doc_content or any(kw in doc_content.lower() for kw in ("chatgpt", "factscore", "dataset viewer")):
                        doc_content = ""
                        index_available.append(i)
                        continue

                    if self.use_in_memory_vectorstore:
                        split_doc_content = CHARACTER_SPLITTER.split_text(doc_content)
                        documents = [Document(id=f"{doc_id}", page_content=chunk, metadata={"source": link})
                            for doc_id, chunk in enumerate(split_doc_content)
                        ]
                        self.in_memory_vectorstore.add_documents(documents=documents)
                        retriever = self.in_memory_vectorstore.as_retriever(search_kwargs={"k": 3})
                        retrieved_docs = retriever.invoke(query_text)
                        doc_content = "\n\n".join([doc.page_content for doc in retrieved_docs])

                    count_content += 1
                else:
                    doc_content = ""
                    count_content += 1

                passages.append(dict(title=title, text=doc_content, snippet=snippet, link=link))

            # --- Fallback to empty-content entries ---
            if count_content < self.top_k:
                for i in index_available:
                    if count_content >= self.top_k:
                        break
                    hit = search_hits[i]
                    passages.append(dict(title=hit["title"], text="", snippet=hit["snippet"], link=hit["link"]))
                    count_content += 1

            results.extend(passages)
            logger.info(f"Retrieved {len(results)} results for query.")
        return results
           

class ContextRetriever:
    """
    Parallel context retriever that wraps a ContextRetriever and dispatches
    retrieval tasks across a thread pool.
    """

    def __init__(
        self,
        retriever: Retriever,
        context_summarizer: Optional[ContextSummarizer] = None,
        num_workers: int = 4,
    ):
        self.retriever = retriever
        self.context_summarizer = context_summarizer
        self.num_workers = num_workers

    def _retrieve_for_item(
        self,
        text: str,
        atom: Optional[Atom] = None,
        id_prefix: str = "c",
    ) -> List[Context]:
        """Worker function: retrieve contexts (and optionally summarize) for one item."""
        retrieved = self.retriever.query(text=text)

        contexts = []
        for j, ctx in enumerate(retrieved):
            context = Context(
                id=f"{id_prefix}_{j}",
                atom=atom,
                text=ctx["text"],
                title=ctx["title"],
                link=ctx["link"],
                snippet=ctx["snippet"],
            )
            contexts.append(context)

        # Summarize in the same worker thread if summarizer is provided
        if self.context_summarizer is not None and atom is not None and len(contexts) > 0:
            results = asyncio.run(
                self.context_summarizer.run_batch(
                    [c.get_text() for c in contexts], atom.text
                )
            )
            for context, result in zip(contexts, results):
                if result["summary"]:
                    context.set_synthetic_summary(result["summary"])
                    context.set_probability(result["probability"] * context.get_probability())
            # Filter out irrelevant contexts (empty summary means not relevant)
            contexts = [c for c in contexts if c.get_summary()]

        return contexts

    def retrieve_all(
        self,
        atoms: Dict[str, Atom],
        query: Optional[str] = None,
    ) -> Dict[str, Context]:
        """Retrieve contexts for all atoms (and optionally the query) in parallel."""
        all_contexts: Dict[str, Context] = {}
        futures = {}

        with ThreadPoolExecutor(max_workers=self.num_workers) as executor:
            # Submit one task per atom
            for aid, atom in atoms.items():
                future = executor.submit(
                    self._retrieve_for_item, atom.text, atom, f"c_{aid}"
                )
                futures[future] = (aid, atom)

            # Submit query task
            if query:
                future = executor.submit(
                    self._retrieve_for_item, query, None, "c_q"
                )
                futures[future] = ("query", None)

            # Collect results with progress bar
            for future in tqdm(
                as_completed(futures),
                total=len(futures),
                desc="Retrieving contexts",
            ):
                aid, atom = futures[future]
                contexts = future.result()
                for ctx in contexts:
                    all_contexts[ctx.id] = ctx
                if atom is not None:
                    atom.add_contexts(contexts)

        return all_contexts



# coding=utf-8
# Copyright 2023-present the International Business Machines.g
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import combinations
from typing import Dict, List, Tuple, Union

import asyncio
import nltk
from nltk.tokenize import sent_tokenize

# Local imports
from .base import Atom, Context, Relation
from .atomizer import Atomizer
from .retriever import ContextRetriever
from .nli import NLIExtractor
from fact_reasoner.utils import punctuation_only_inside_quotes

def predict_nli_relationships(
        object_pairs: List[Tuple[Union[Atom, Context], Union[Atom, Context]]],
        nli_extractor: NLIExtractor,
        links_type: str = "context_atom",
        use_summary: bool = False,
    ) -> List[Relation]:
    """
    Predict the NLI relationship between two objects using an model based NLI extractor.

    Args:
        object_pairs: List
            A list of object pairs e.g., (atom, context) or (context, context)
        nli_extractor: NLIExtractor
            The model based NLI extractor
        top_k_per_atom: int
            The top k relationships considered for each atom.
        links_type: str
            The type of links represented by the object pairs (context_atom, context_context).
    """

    # Safety checks
    assert (nli_extractor is not None), "NLI extractor cannot be None."
    assert isinstance(nli_extractor, NLIExtractor), "NLI extractor must be NLIExtractor."

    # Set up the premises and hypotheses
    if use_summary:
        premises = [pair[0] if isinstance(pair[0], str) else pair[0].get_summary() for pair in object_pairs]
        hypotheses = [pair[1] if isinstance(pair[1], str) else pair[1].get_summary() for pair in object_pairs]
    else:
        premises = [pair[0] if isinstance(pair[0],str) else pair[0].get_text() for pair in object_pairs]
        hypotheses = [pair[1] if isinstance(pair[1],str) else pair[1].get_text() for pair in object_pairs]

    # Safety checks
    assert len(premises) == len(hypotheses)

    # Extract the NLI relationships between premises and hyptheses
    print(f"[NLI] Processing {len(premises)} potential relationships ...")
    # results = [nli_extractor.run(premises[i], hypotheses[i]) for i in range(len(premises))]
    results = asyncio.run(nli_extractor.run_batch(premises, hypotheses))

    relations = []
    for ii, result in enumerate(results):
        label = result.get("label", "neutral")
        probability = result.get("probability", 0.0)
        link_type = links_type if links_type is not None else "unknown"
        rel = Relation(
            source=object_pairs[ii][0],
            target=object_pairs[ii][1],
            type=label,
            probability=probability,
            link=link_type
        )
        relations.append(rel)
   
    return relations

def build_atoms(response: str, atom_extractor: Atomizer) -> Dict[str, Atom]:
    """
    Decompose the given response into atomic units (i.e., atoms).

    Args:
        response: str
            The string representing the LLM response.
        atom_extractor: Atomizer 
            The atom extractor.

    Returns:
        Dict[str, Atom]: A dict containing the atoms of the response.
    """

    assert (response is not None and len(response) > 0), \
        f"Please ensure a non empty response."

    result = atom_extractor.run(response)
    candidates = [
        Atom(
            id="a" + str(i),
            text=v
        ) for i, v in enumerate(result.values())
    ]

    return {atom.id: atom for atom in candidates}


def build_contexts(
        atoms: Dict[str, Atom] = {},
        query: str = None,
        retriever: ContextRetriever = None,
        use_fast_retriever: bool = True
) -> Dict[str, Context]:
    """
    Retrieve the relevant contexts for the input atoms.

    Args:
        atoms: dict
            A dict containing the atoms in the response.
        query: str
            The user query text.
        retriever: ContextRetriever 
            The context retriever (chromadb, langchain, google).
        use_fast_retriever: bool
            Use the fast multi-threaded context retriever.
    
    Returns:
        Dict[str, Context]: A dict containing the retrieved contexts.
    """

    assert (len(atoms) > 0), \
        "Please ensure a non-empty list of atoms."
    assert (retriever is not None), \
        "Please ensure an existing context retriever instance."

    # Building the contexts
    contexts = {}

    if not use_fast_retriever:
        # Retrieve contexts for the atoms
        for aid, atom in atoms.items():
            
            # Sequential but with multi-threaded top-k retrieval
            retrieved_contexts = retriever.context_retriever.query(
                text=atom.text,
            )
            
            if len(retrieved_contexts) > 0:
                contexts_per_atom = [
                    Context(
                        id="c_" + aid + "_" + str(j),
                        atom=atom,
                        text=context["text"],
                        title=context["title"],
                        link=context["link"],
                        snippet=context["snippet"]
                        # An empty summary means that the context is not relevant, 
                        # therefore we do not add it to the list of contexts for the pipeline
                    ) for j, context in enumerate(retrieved_contexts) 
                ]

                for ctxt in contexts_per_atom:
                    contexts[ctxt.id] = ctxt
                atoms[aid].add_contexts(contexts_per_atom)

        # Retrieve the contexts for the question
        retrieved_contexts = retriever.context_retriever.query(
            text=query,
        )
    
        if len(retrieved_contexts) > 0:
            contexts_per_query = [
                Context(
                    id="c_q_" + str(j),
                    atom=None,
                    text=context["text"],
                    title=context["title"],
                    link=context["link"],
                    snippet=context["snippet"]
                    # An empty summary means that the context is not relevant, 
                    # therefore we do not add it to the list of contexts for the pipeline
                ) for j, context in enumerate(retrieved_contexts) 
            ]

            for ctxt in contexts_per_query:
                contexts[ctxt.id] = ctxt
    else:
        # Retrieve contexts for all atoms in parallel
        contexts = retriever.retrieve_all(atoms=atoms, query=query)
    
    return contexts

def remove_duplicated_atoms(atoms: Dict[str, Atom]) -> Dict[str, Atom]:
    """
    Remove the duplicated atoms.

    Args:
        atoms: Dict[str, Any]
            The dict containing the atoms.

    Returns:
        Dict[str, Any]: A dict containing the unique atoms.
    """

    seen = set()
    out = {}
    for k, v in atoms.items():
        text = v.get_text()
        if text not in seen:
            out[k] = v
            seen.add(text)
    return out


def remove_duplicated_contexts(contexts: Dict[str, Context], atoms: Dict[str, Atom], check_summary: bool = False) -> dict:
    """
    Remove the duplicated contexts.
    
    Args:
        contexts: Dict[str, Context]
            The dict containing the contexts.
        atoms: Dict[str, Atom]
            The dict containing the atoms.
        check_summary: bool
            Whether to check the summary of the contexts.

    Returns:
        The updated dicts containing the contexts and atoms.
    """

    seen = set()
    out = {}
    for k, v in contexts.items():
        text = v.get_text() if not check_summary or v.get_summary() is None else v.get_summary()
        if text not in seen:
            seen.add(text)
            out[k] = v
        elif v.atom and v.atom.id in atoms:
            del atoms[v.atom.id].contexts[k]
    
    return out, atoms

def is_relevant_context(context: str) -> bool:
    """
    Check if context is relevant.
    """

    keywords = [
        "not provide information about the atom",
        "not provide any information about the atom",
        "not provide specific information about the atom",
        "not contain information about the atom",
        
        "not provide any information related to the atom",
        "not provide specific information related to the atom",
        "not provide information related to the atom",
        
        "not contain information about the atom",   
        "not contain any information about the atom",
        "not contain specific information about the atom",
        
        "not provide information on the atom",
        "not provide any information on the atom",
        "not provide specific information on the atom",
        
        "insufficient to make a conclusion about the atom",
        "not provide enough information to make a conclusion about the atom",
        "not contain enough information to make a conclusion about the atom",

        "not provide any relevant information about the atom",
        "information about the atom cannot be found",
        "information is not about the atom",
        "information is not related to the atom",

        "is known that",
        "is generally known that",
        "is believed that",
        
        "don't have permission to view this page",
        "due to a 403 forbidden error",
        "shows a 403 forbidden error",
        "is a 403 forbidden error",
        "not have permission to view",
        "not have permission to access",
        "access to the page is forbidden",
        "context is not available",
        "context is not accessible",

        "not possible to summarize the context",
        "verify the given atom",

        "atom statement",
        "atom states",
    ]

    context_lower = context.lower()
    if not all(keyword.lower() not in context_lower for keyword in keywords):
        return False
            
    try:
        nltk.data.find('tokenizers/punkt')
    except LookupError:
        print("'punkt' not found. Downloading...")
        nltk.download('punkt')

    sentences = sent_tokenize(context)
    num_sentences = len(sentences)
    # we filter out summaries of only one sentence of the form: "the context does not..."
    if (num_sentences == 1 
        and punctuation_only_inside_quotes(sentences[0]) 
        and ("the context does not" in sentences[0].lower())
       ):
            return False
    
    return True


def build_relations(
        atoms: Dict[str, Atom] = {},
        contexts: Dict[str, Context] = {},
        contexts_per_atom_only: bool = False,
        rel_atom_context: bool = True, 
        rel_context_context: bool = True,
        nli_extractor: NLIExtractor = None,
        use_summarized_contexts: bool = False
) -> List[Relation]:
    """
    Create the NLI relations between atoms and contexts. The following
    pairwise relations are considered: atom-context and context-context.

    Args:
        atoms: dict
            A dict containing the atoms in the response.
        contexts: dict
            A dict containing the contexts retrived from the vector store.
        contexts_per_atom_only: bool
            Flag indicating that for each atom only its corresponding contexts are considered. 
        rel_atom_context: bool (default is True)
            Flag indicating the presence of atom-to-context relationships.
        rel_context_context: bool (default is False)
            Flag indicating the presence of context-to-context relationships.
        nli_extractor: NLIExtractor
            The NLI model used for predicting the relationships.
        use_summarized_contexts: bool
            Flag indicating that summarized contexts are used. If False, then the
            contexts include the extracted text.
    Returns:
        A list of Relations.  
    """

    assert len(atoms) > 0, f"The atoms must be initialized!"
    assert len(contexts) > 0, f"The contexts must be initialized!"
    assert (nli_extractor is not None), f"The NLI extractor must exist!"
    
    atom_context_pairs = []
    context_context_pairs1 = []
    context_context_pairs2 = []

    relations = []

    # Create atom-context relations (i.e., Context -> Atom)
    if rel_atom_context:
        print(f"[NLI] Building atom-context relations...")
        if not contexts_per_atom_only:  # use all contexts for each atom
            # Create the (context, atom) pairs
            print(f"[NLI] Using all contexts retrieved.")
            for _, atom in atoms.items():
                for _, context in contexts.items():
                    atom_context_pairs.append((context, atom))
        else:
            print(f"[NLI] Using only the contexts retrieved per atom.")
            # Create the (context, atom) pairs
            for _, atom in atoms.items():
                for context in atom.get_contexts():
                    atom_context_pairs.append((context, atom))

        # Get all relationships (NLI-prompt)
        all_rels = predict_nli_relationships(
            atom_context_pairs,
            nli_extractor=nli_extractor,
            links_type="context_atom",
            use_summary=use_summarized_contexts
        )

        # Filter out the neutral relationships
        for rel in all_rels:
            if rel.get_type() != "neutral":
                print(f"[NLI] Found relation: {rel}")
                relations.append(rel)

    # Create context-context relations
    if rel_context_context:
        print(f"[NLI] Building context-context relations...")
        clist = [ci for ci in sorted(contexts.keys())]
        all_pairs = list(combinations(clist, 2))
        # Create all (context, context) pairs
        for ci, cj in all_pairs:
            context_i = contexts[ci]
            context_j = contexts[cj]
            context_context_pairs1.append((context_i, context_j))
            context_context_pairs2.append((context_j, context_i))

        # Get relationships (c_i, c_j)
        relations1 = predict_nli_relationships(
            context_context_pairs1,
            nli_extractor=nli_extractor,
            links_type="context_context",
            use_summary=use_summarized_contexts
        )

        # Get relationships (c_j, c_i)
        relations2 = predict_nli_relationships(
            context_context_pairs2,
            nli_extractor=nli_extractor,
            links_type="context_context",
            use_summary=use_summarized_contexts
        )

        relations_tmp = [pair[0] if pair[0].get_probability()>pair[1].get_probability() else pair[1] for pair in zip(relations1,relations2)]
        assert len(relations_tmp) == len(relations1) # safety checks

        for rel_ind in range(len(relations_tmp)):
            if not (relations1[rel_ind].get_type() == "entailment" and relations2[
                rel_ind].get_type() == "entailment"): continue
            relations_tmp[rel_ind].type = "equivalence"
        for rel in relations_tmp:
            if rel.get_type() != "neutral":
                print(f"[NLI] Found relation: {rel}")
                relations.append(rel)

    print(f"[NLI] Relations built: {len(relations)}")
    return relations



# coding=utf-8
# Copyright 2023-present the International Business Machines.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Our implementation of the SAFE paper with LLAMA3 models where facts/claims
# are checked against Google search results.
 
import json
import asyncio
import time
import mellea.stdlib.functional as mfuncs
from typing import List, Dict, Any, Tuple

from mellea.backends import Backend
from mellea.stdlib.context import SimpleContext
from mellea.core import ModelOutputThunk
from mellea.stdlib.requirements import check
from mellea.stdlib.sampling import RejectionSamplingStrategy
from mellea.core import FancyLogger

# Local imports
from fact_reasoner.core.atomizer import Atomizer
from fact_reasoner.core.reviser import Reviser
from fact_reasoner.core.retriever import ContextRetriever
from fact_reasoner.core.base import Atom, Context
from fact_reasoner.core.utils import build_atoms, build_contexts, remove_duplicated_atoms
from fact_reasoner.utils import extract_last_wrapped_response, LOOP_BUDGET

INSTRUCTION_FACTVERIFY = """

You need to judge whether a claim is supported or contradicted by Google search results, or whether there is no enough information to make the judgement. When doing the task, take into consideration whether the link of the search result is of a trustworthy source. Mark your answer with ### signs.

Below are the definitions of the three categories:

Supported: A claim is supported by the search results if everything in the claim is supported and nothing is contradicted by the search results. There can be some search results that are not fully related to the claim.
Contradicted: A claim is contradicted by the search results if something in the claim is contradicted by some search results. There should be no search result that supports the same part.
Inconclusive: A claim is inconclusive based on the search results if:
- a part of a claim cannot be verified by the search results,
- a part of a claim is supported and contradicted by different pieces of evidence,
- the entity/person mentioned in the claim has no clear referent (e.g., "the approach", "Emily", "a book").

Here are some examples:

Claim: Characters Lenny and Carl on The Simpsons are hearing but are depicted as close friends of the Simpsons family.

Search result 1\nTitle: Character Spotlight: Lenny Leonard and Carl Carlson (& Barflies)\nContent: Their friendship is a pretty singular aspect on the show -- save Bart and Milhouse (or to some degree, Mr. Burns and Smithers) -- they always ...\nLink: https://nohomers.net/forums/index.php?threads/character-spotlight-lenny-leonard-and-carl-carlson-barflies.23798/\n\nSearch result 2\nTitle: The Simpsons: Lenny and Carl's History, Explained - CBR\nContent: Introduced in the show's first season, the pair were portrayed as background characters at Homer's work, usually appearing together in minor ...\nLink: https://www.cbr.com/the-simpsons-lenny-carl-history-explained/\n\nSearch result 3\nTitle: Are Lennie and Carl Homer Simpson's real or fake friends? - Quora\nContent: Lenni is a pal, Carl doesn't consider any of them to be 'friends' they're just shallow guys he hangs out with. Lenny and Carl have a special ...\nLink: https://www.quora.com/Are-Lennie-and-Carl-Homer-Simpson-s-real-or-fake-friends\n\nSearch result 4\nTitle: [The Simpsons] Lenny and Carl aren't ambiguously gay (originally)\nContent: Theory: Lenny and Carl started out as a parody of background characters who always appear as a pair: Crabbe and Goyle or Fred and George in \" ...\nLink: https://www.reddit.com/r/FanTheories/comments/yw7bp4/the_simpsons_lenny_and_carl_arent_ambiguously_gay/\n\nSearch result 5\nTitle: Lenny Leonard | Simpsons Wiki | Fandom\nContent: He is the best friend of Carl Carlson and, along with Carl, second best friend of Homer Simpson, behind Barney Gumble and Moe Syzslak.\nLink: https://simpsons.fandom.com/wiki/Lenny_Leonard\n\nSearch result 6\nTitle: Are Simpsons' Carl & Lenny Gay? Every Clue To Their Relationship\nContent: One of The Simpsons' many mysteries is Lenny and Carl's relationship, as it has been hinted that the two are more than just best friends.\nLink: https://screenrant.com/simpsons-shows-carl-lenny-gay-couple-clues-hints/\n\nSearch result 7\nTitle: Lenny Leonard - Wikisimpsons, the Simpsons Wiki\nContent: Lenford \"Lenny\" Leonard is the best friend of Carl Carlson and, along with Carl, second best friend of Homer Simpson, behind Barney Gumble.\nLink: https://simpsonswiki.com/wiki/Lenny_Leonard\n\nSearch result 8\nTitle: Lenny | The Simpsons: Tapped Out Wiki | Fandom\nContent: Lenford \"Lenny\" Leonard MPhs (born 1960) is the best friend of Carl Carlson and, along with Carl, second best friend of Homer Simpson, behind Barney Gumble.\nLink: https://simpsonstappedout.fandom.com/wiki/Lenny\n\nSearch result 9\nTitle: The Simpsons - Wikipedia\nContent: Developed by Groening, James L. Brooks, and Sam Simon, the series is a satirical depiction of American life, epitomized by the Simpson family, which consists of ...\nLink: https://en.wikipedia.org/wiki/The_Simpsons\n\nSearch result 10\nTitle: Lenny Leonard & Carl Carlson - Friends or Couple? | The Simpsons\nContent: Embark on an Epic Friendship Adventure: Lenny Leonard and Carl Carlson from The Simpsons ...\nLink: https://www.youtube.com/watch?v=qY5hjalUhfA"

Your decision: ###Inconclusive###

Claim: The championship match of the FIFA World Cup 2026 will be hosted by the United States.

"Search result 1\nTitle: World Cup 2026 | New York New Jersey to host final - FIFA\nContent: New York New Jersey Stadium has been confirmed as the location for the FIFA World Cup 26™ final on Sunday, 19 July 2026. The full match schedule for the ...\nLink: https://www.fifa.com/fifaplus/en/tournaments/mens/worldcup/canadamexicousa2026/articles/new-york-new-jersey-stadium-host-world-cup-2026-final\n\nSearch result 2\nTitle: 2026 FIFA World Cup - Wikipedia\nContent: The tournament will take place from June 11 to July 19, 2026. It will be jointly hosted by 16 cities in three North American countries: Canada, Mexico, and the ...\nLink: https://en.wikipedia.org/wiki/2026_FIFA_World_Cup\n\nSearch result 3\nTitle: World Cup 2026 | Dallas to host nine matches - FIFA\nContent: Dallas Stadium will host nine matches from the FIFA World Cup 26™, including four knockout games in the latter stages of the tournament.\nLink: https://www.fifa.com/fifaplus/en/tournaments/mens/worldcup/canadamexicousa2026/articles/dallas-stadium-host-nine-world-cup-matches\n\nSearch result 4\nTitle: When and where is the 2026 World Cup? What is the format?\nContent: 104 games will take place in the 2026 World Cup across the USA, Mexico, and Canada with 16 different host cities selected and the schedule confirmed.\nLink: https://www.nbcsports.com/soccer/news/when-and-where-is-the-2026-world-cup\n\nSearch result 5\nTitle: New York/New Jersey will host 2026 FIFA World Cup final\nContent: FIFA announced on Sunday that the final of the 2026 World Cup will be held at MetLife Stadium just outside New York.\nLink: https://www.foxsports.com/stories/soccer/new-york-new-jersey-will-host-2026-fifa-world-cup-final-mexico-city-gets-the-opener\n\nSearch result 6\nTitle: FIFA announces 2026 World Cup details, including host stadium for ...\nContent: The tournament will be hosted by the United States, Mexico and Canada. PHOTO: A general exterior view of the MetLife Stadium the home of NFL New ...\nLink: https://abcnews.go.com/Sports/new-jersey-metlife-stadium-host-2026-world-cup-final/story?id=106937005\n\nSearch result 7\nTitle: New York, New Jersey to host 2026 FIFA World Cup final - France 24\nContent: The 2026 World Cup final will be held at MetLife Stadium in New York/New Jersey, organisers FIFA announced on Sunday.\nLink: https://www.france24.com/en/americas/20240204-new-york-new-jersey-to-host-2026-fifa-world-cup-final\n\nSearch result 8\nTitle: 2026 World Cup final will be played at MetLife Stadium in New Jersey\nContent: The 2026 World Cup final will be played at MetLife Stadium in New Jersey, beating out Texas and California for soccer's showcase game.\nLink: https://www.nbcmiami.com/news/sports/miami-misses-out-on-hosting-the-2026-fifa-world-cup-final-metlife-stadium-in-new-jersey-to-host/3224690/\n\nSearch result 9\nTitle: New Jersey's MetLife Stadium to host 2026 FIFA World Cup final on ...\nContent: The 104-match tournament will open in Mexico on June 11 and will move entirely to the US from the quarterfinal round. MetLife Stadium in East ...\nLink: https://www.aljazeera.com/sports/2024/2/5/new-jersey-to-host-2026-fifa-world-cup-final\n\nSearch result 10\nTitle: World Cup 2026 | Seattle to host six matches - FIFA\nContent: Seattle Stadium will host six matches from the FIFA World Cup 26™, including USA's second group fixture and two knockout games. The full match schedule for ...\nLink: https://www.fifa.com/fifaplus/en/tournaments/mens/worldcup/canadamexicousa2026/articles/seattle-host-six-matches-stadium

Your decision: ###Supported###

Claim: It is essential to understand the limitations of heating a dead battery to temporarily revive its function.

"Search result 1\nTitle: Why do batteries come back to life if you let them rest?\nContent: By letting the battery rest, you give the reaction products a chance to dissipate. The higher the drain on the battery, the faster the products ...\nLink: https://electronics.howstuffworks.com/everyday-tech/question390.htm\n\nSearch result 2\nTitle: ELI5 Why do batteries work again after heating them up? - Reddit\nContent: The warmth heats up the battery and give it a little more juice allowing me to use it again for a minute.\nLink: https://www.reddit.com/r/explainlikeimfive/comments/vjn30b/eli5_why_do_batteries_work_again_after_heating/\n\nSearch result 3\nTitle: What happen when a dead battery and a good battery connected in ...\nContent: The batteries could get warm/hot/overheat and or get damaged depending on the capacity of the batteries being connected and their ability to ...\nLink: https://www.quora.com/What-happen-when-a-dead-battery-and-a-good-battery-connected-in-parallel\n\nSearch result 4\nTitle: The flat battery vs dead battery - key differences and how to handle t ...\nContent: A flat battery is temporarily depleted but can be recharged and regain its functionality. ... its ability to hold a charge and cannot be revived.\nLink: https://www.tycorun.com/blogs/news/flat-battery-vs-dead-battery\n\nSearch result 5\nTitle: How to Bring Back 0V/low voltage NiMH Battery to Life – EBLOfficial\nContent: In such cases, it can be essential to overcharge your NiMH battery sometimes to ensure they are fully charged and operate optimally. However, ...\nLink: https://www.eblofficial.com/blogs/blog/how-to-revive-nimh-battery\n\nSearch result 6\nTitle: How Long Should You Drive to Charge Your Car Battery?\nContent: Discover how long you should drive to charge your car battery effectively. Learn tips, calculations, and alternative methods. Drive smart, charge smart!\nLink: https://heartautocare.com/how-long-should-you-drive-to-charge-your-car-battery/\n\nSearch result 7\nTitle: BU-502: Discharging at High and Low Temperatures\nContent: Like humans, batteries function best at room temperature. Warming a dying battery in a mobile phone or flashlight in our jeans might provide ...\nLink: https://batteryuniversity.com/article/bu-502-discharging-at-high-and-low-temperatures\n\nSearch result 8\nTitle: Can You Recharge A Completely Dead Car Battery?\nContent: Learn the signs of a dead battery, whether you can recharge a completely dead car battery and other method to safely revive it.\nLink: https://carfromjapan.com/article/car-maintenance/can-you-recharge-a-completely-dead-car-battery/\n\nSearch result 9\nTitle: How to Fix a Dead Car Battery - Naylor's Auto Repair\nContent: It indicates that the battery is leaking. This will tell that your battery will not function as it should.\nLink: https://www.naylorsautorepairidaho.com/blog/how-fix-dead-car-battery\n\nSearch result 10\nTitle: The Truth About Reviving Dead Batteries - U.S. Battery Mfg. Co.\nContent: The batteries may or may not recover after one desulfation charge. If not, it may require several cycles of discharge and recharge with ...\nLink: https://www.usbattery.com/the-truth-about-reviving-dead-batteries/

Your decision: ###Inconclusive###

Claim: Sarah and James were shot execution-style in the living room.

"Search result 1\nTitle: By Taufik | There were string of armed robberies and free murders ...\nContent: handgun been in the execution-style murder but we ... quickly cleared the living room, went ...\nLink: https://www.facebook.com/Haru6789.cv/videos/watch-killer-siblings-season-3-episode-5-allridges/411826944038089/\n\nSearch result 2\nTitle: Why are Austin's Yogurt Shop Murders Still Unsolved After 30 Years?\nContent: Thomas, Ayers, and Jennifer and Sarah Harbinson were all shot and killed execution-style with a single gunshot wound to the back of their heads.\nLink: https://www.oxygen.com/crime-news/yogurt-shop-murders-austin-solved-theories-dna-update\n\nSearch result 3\nTitle: Jenean Chapman's neighbors say they heard former aide to Sarah ...\nContent: Neighbors revealed they heard 'screaming fights' between Sarah Ferguson's ex-assistant and her husband, James Patrick, who has been arrested ...\nLink: https://www.dailymail.co.uk/news/article-12582157/Neighbors-Sarah-Ferguson-fights-James-Patrick-Jenean-Chapman-Dallas-Texas.html\n\nSearch result 4\nTitle: Sarah Brady, gun control advocate and widow of James Brady, dies ...\nContent: Sarah Brady, gun control advocate and widow of James Brady, dies at 73 ... After the 1981 assassination attempt on President Ronald Reagan, nearly ...\nLink: https://www.latimes.com/local/obituaries/la-me-sarah-brady-20150403-story.html\n\nSearch result 5\nTitle: Slayings of 3 Youths Stun Little Rock - The New York Times\nContent: When three armed men invaded the home of Mary Hussian here this month and killed three of her children, execution-style, this city, ...\nLink: https://www.nytimes.com/1995/06/27/us/slayings-of-3-youths-stun-little-rock.html\n\nSearch result 6\nTitle: Revenge Is Motive in 7 Bronx Slayings, Police Say - The New York ...\nContent: Valentine's Day, they said, when three adults and three teen-agers were lined up on the living-room floor of a Bronx apartment and each shot in ...\nLink: https://www.nytimes.com/1993/02/26/nyregion/revenge-is-motive-in-7-bronx-slayings-police-say.html\n\nSearch result 7\nTitle: 20 years later, triple murder casts a shadow over Christmas ...\nContent: The grisly, Christmas Day discovery of a Jean Wholaver and her two daughters shot dead execution-style by their ex-husband and father, Ernest ...\nLink: https://www.pennlive.com/crime/2022/12/20-years-later-triple-murder-casts-a-shadow-over-christmas-memories-in-middletown.html\n\nSearch result 8\nTitle: Murder victims identified after police say they were shot 'execution ...\nContent: Murder victims identified after police say they were shot 'execution style'. 7.6K views · 1 ...\nLink: https://www.youtube.com/watch?v=ux9L1NLm-BI\n\nSearch result 9\nTitle: Bonus – the James Bigby murders - what was that like\nContent: The former auto mechanic is accused of shooting three former coworkers and strangling a 4-month-old baby. We have more in this NBC report. The 4 ...\nLink: https://whatwasthatlike.com/2023/04/14/bonus-the-james-bigby-murders/\n\nSearch result 10\nTitle: Man arrested over 'execution-style' killings in Philadelphia park\nContent: A man has been arrested over the \"execution-style\" killings of a man and woman in a Philadelphia park. The bodies of Thurston Cooper, 49, ...\nLink: https://www.irishstar.com/news/pennsylvania-news/execution-style-killings-philadelphia-park-32278732

Your decision: ###Inconclusive###

Claim: Vikings used their longships to transport livestock.

Search result 1\nTitle: How did the Vikings transport animals on their ships? - Quora\nContent: The Vikings transported horses overseas in boats very similar to Viking longships, but with flat flooring built within the hulls, which allowed ...\nLink: https://www.quora.com/How-did-the-Vikings-transport-animals-on-their-ships\n\nSearch result 2\nTitle: The Truth Behind Vikings Ships\nContent: They could land on any beach, permitting lightning-quick embarking and attacks. Great loads could be carried, including horses and livestock.\nLink: https://www.vikings.com/news/the-truth-behind-vikings-ships-18274806\n\nSearch result 3\nTitle: Viking ships | Royal Museums Greenwich\nContent: Cargo vessels were used to carry trade goods and possessions. They were wider than the longships and travelled more slowly.\nLink: https://www.rmg.co.uk/stories/topics/viking-ships\n\nSearch result 4\nTitle: How did the vikings bring the horses to Iceland? - Reddit\nContent: This was stored in casks and skins. It's worth mentioning that some discomfort for these animals was expected, and they may have been 'rationed' ...\nLink: https://www.reddit.com/r/AskHistorians/comments/5l05tj/how_did_the_vikings_bring_the_horses_to_iceland/\n\nSearch result 5\nTitle: Did the Vikings bring livestock on their raids and voyages?\nContent: The Vikings did not generally transport livestock during their raids. Viking longships did not have substantial cargo room. As a result, the Vikings only ...\nLink: https://homework.study.com/explanation/did-the-vikings-bring-livestock-on-their-raids-and-voyages.html\n\nSearch result 6\nTitle: LEVS : Viking FAQs : Transportation\nContent: The upper deck was used for humans and their space, and the lower decks were used for livestock, furniture, and other goods that were being transported from ...\nLink: http://vikingship.org/ourfaqs/transportation_1.html\n\nSearch result 7\nTitle: Viking Ships - World History Encyclopedia\nContent: Such ships gave the Vikings the ability to trade, make war, carry animals ... Longships, on the other hand, used both oars and sails to reach ...\nLink: https://www.worldhistory.org/Viking_Ships/\n\nSearch result 8\nTitle: The Viking Longship - Warfare History Network\nContent: Longships enabled the Vikings to transport their armies throughout Europe and conduct amphibious assaults in estuaries and navigable rivers.\nLink: https://warfarehistorynetwork.com/article/the-viking-longship/\n\nSearch result 9\nTitle: Viking Ships | Regia Anglorum\nContent: Most Viking trading boats were able to come ashore without the need of complicated harbour facilities, and yet carry many tons of cargo. This encouraged ...\nLink: https://regia.org/research/ships/Ships0.htm

Your decision: ###Contradicted###

Claim: Romário has scored a total of 92 international goals.

Search result 1\nTitle: Romário - Wikipedia\nContent: A prolific striker renowned for his clinical finishing, he scored over 700 goals and is one of the few players to score at least 100 goals for three clubs. He ...\nLink: https://en.wikipedia.org/wiki/Rom%C3%A1rio\n\nSearch result 2\nTitle: Romário de Souza Faria - Goals in International Matches - RSSSF\nContent: The outstanding Brazilian player, who scored 55 international goals for his National Team, has ... 92 Paris France 2-0 27 10 16/12/92 Porto ...\nLink: https://www.rsssf.org/miscellaneous/romario-intlg.html\n\nSearch result 3\nTitle: In 92 Romário said:\" I didn't come from Europe for a friendly in Brazil ...\nContent: he played till he was like 42, scored 772 goals in official games in his career (the same average as messi and higher than ronaldo) and won a ...\nLink: https://www.reddit.com/r/soccer/comments/7klck5/in_92_rom%C3%A1rio_said_i_didnt_come_from_europe_for_a/\n\nSearch result 4\nTitle: Romário - National team - Transfermarkt\nContent: Only games with goals scored. Compact · Detailed. Matchday, Date, Match, Pos. International Friendlies. 5/23/87, Republic of Ireland · 1:0 ...\nLink: https://www.transfermarkt.us/romario/nationalmannschaft/spieler/7942\n\nSearch result 5\nTitle: List of men's footballers with 50 or more international goals - Wikipedia\nContent: Cristiano Ronaldo holds the all-time record with 128 goals. Brazil and Hungary hold the record of having the most players to have scored 50 or more ...\nLink: https://en.wikipedia.org/wiki/List_of_men%27s_footballers_with_50_or_more_international_goals\n\nSearch result 6\nTitle: romário de souza faria - Romàrio - FC Barcelona Players\nContent: According to recent studies, he is the highest goal scorer in world football, with 768 goals in official matches for club and national team, ahead of Bican, ...\nLink: https://players.fcbarcelona.com/en/player/759-romario-romario-souza-faria\n\nSearch result 7\nTitle: Romario is the best striker ever imo | BigSoccer Forum\nContent: * Dutch league's top scorer: 1989, 1990, 1991, 1992 ... romario has scored 55 goals , thats about 34 % of ... in 1994 world cup, in brazil games was scored a total ...\nLink: https://www.bigsoccer.com/threads/romario-is-the-best-striker-ever-imo.597679/\n\nSearch result 8\nTitle: Romario: The Art of Goal-Scoring Brilliance. - LinkedIn\nContent: In 1997 alone, the duo scored an impressive total of 34 international goals with 19 coming from Romário. The Ro-Ro attack was expected to ...\nLink: https://www.linkedin.com/pulse/romario-art-goal-scoring-brilliance-javad-ghorbani\n\nSearch result 9\nTitle: Who is Brazil's leading all-time top goal scorer? Pele, Neymar ...\nContent: The three-time World Cup winner went on to score a total of 77 goals in 92 games for the Selecao at an astonishing 0.84 goals-per-game ratio.\nLink: https://www.goal.com/en-us/lists/pele-neymar-ronaldo-who-is-brazils-leading-all-time-top-goalscorers/blt75e4aa20fd6a2bed\n\nSearch result 10\nTitle: Romário - All goals - Transfermarkt\nContent: This overview shows all goals scored by the selected player, including the results of the games. It is also possible to select a competition. Filter by season:.\nLink: https://www.transfermarkt.us/romario/alletore/spieler/7942

Your decision: ###Contradicted###

Claim: Utopia portrays a society that values education and learning.

Search result 1\nTitle: Utopia Education, Science, Philosophy Summary & Analysis\nContent: The Utopians believe that it is through education that the values and dispositions of citizens are molded. The success of the Utopian educational system is ...\nLink: https://www.sparknotes.com/philosophy/utopia/section10/\n\nSearch result 2\nTitle: Utopianism and Education: The Legacy of Thomas More - jstor\nContent: Utopia, are filled from this sector of society. ... fail to find significance and meaning in classroom learning. ... closed, value positions about education - about ...\nLink: https://www.jstor.org/stable/3122242\n\nSearch result 3\nTitle: [PDF] education in thomas more's utopia - DergiPark\nContent: This work of art depicts an ideal society and in doing so relays many qualities desirable in an educational system. This study aims to explore Thomas More's ...\nLink: https://dergipark.org.tr/tr/download/article-file/55667\n\nSearch result 4\nTitle: Society and Government in Thomas More's “Utopia” - Medium\nContent: The utopian society portrayed in More's book employs enslaved people for animal slaughter and heavy labour. However, the discussion of slavery ...\nLink: https://medium.com/@batuhankarakus95/society-and-government-in-thomas-mores-utopia-a-focus-on-values-and-functioning-d9626aeaee71\n\nSearch result 5\nTitle: Utopia by Thomas More | Summary, Characters & Themes - Study.com\nContent: According to Thomas More, Utopian society is based on rational thought, communal property, productivity, no class distinctions or poverty, little crime or ...\nLink: https://study.com/academy/lesson/utopia-by-thomas-more-summary-analysis-quiz.html\n\nSearch result 6\nTitle: Thomas More's Utopian Education | Inside Classical Education\nContent: We keep hearing of Utopian visions of culture and society, and I have been itching to go to the sources the word and the concept. More's book ...\nLink: https://insideclassicaled.com/thomas-mores-utopian-education/\n\nSearch result 7\nTitle: About Utopia and Utopian Literature - Cliffs Notes\nContent: Throughout the society, life is directed by a highly moral code of conduct. An educational system for the intelligentsia is elaborately and idealistically ...\nLink: https://www.cliffsnotes.com/literature/u/utopia-utopian-literature/about-utopia-and-utopian-literature\n\nSearch result 8\nTitle: Full article: Educational Studies and the Domestication of Utopia\nContent: Utopia depicts an entire functioning society. Utopian visions 'are explicitly holistic, imaginary, critical, normative, prescriptive ...\nLink: https://www.tandfonline.com/doi/full/10.1080/00071005.2016.1143085\n\nSearch result 9\nTitle: Utopian Society | Definition, Ideas & Examples - Lesson - Study.com\nContent: The idea behind utopianism is a society in which everyone's needs are met and society's ills have been defeated. Because this is an extremely tall order, ...\nLink: https://study.com/academy/lesson/characteristics-of-a-utopian-society.html\n\nSearch result 10\nTitle: Education and Utopia: Robert Owen and Charles Fourier - Jstor\nContent: into the community—his ideal society contains no schools and no teachers. ... (Universities are not mentioned in A new view of society, but are later portrayed as ...\nLink: https://www.jstor.org/stable/23119459

Your decision: ###Supported###

Claim: The higher density of water can cause sound waves to be reflected or refracted differently.

Search result 1\nTitle: How does sound in air differ from sound in water?\nContent: Sounds in water and sounds in air that have the same pressures have very different intensities because the density of water is much greater than ...\nLink: https://dosits.org/science/sounds-in-the-sea/how-does-sound-in-air-differ-from-sound-in-water/\n\nSearch result 2\nTitle: When a sound wave passes from air into water, what properties of ...\nContent: Sound travels faster in water than in air because the density of water is higher. The exact speed depends on the temperature, pressure, and ...\nLink: https://www.quora.com/When-a-sound-wave-passes-from-air-into-water-what-properties-of-the-wave-will-change-1\n\nSearch result 3\nTitle: Reflection, Refraction, and Diffraction - The Physics Classroom\nContent: Sound waves travel slower in cooler air than they do in warmer air. For this reason, the portion of the wavefront directly above the water is slowed down, while ...\nLink: https://www.physicsclassroom.com/class/sound/Lesson-3/Reflection,-Refraction,-and-Diffraction\n\nSearch result 4\nTitle: Refraction of Sound Waves - Graduate Program in Acoustics\nContent: When a wave encounters different medium where the wave speed is different, the wave will change directions. Most often refraction is encountered in a study ...\nLink: https://www.acs.psu.edu/drussell/demos/refract/refract.html\n\nSearch result 5\nTitle: Refraction of light - Science Learning Hub\nContent: When light travels from air into water, it slows down, causing it to change direction slightly. This change of direction is called refraction.\nLink: https://www.sciencelearn.org.nz/resources/49-refraction-of-light\n\nSearch result 6\nTitle: How is sound refracted going from a less dense media to a denser ...\nContent: Refraction is the property of a wave to bend as it propagates through different media. The changes in the media makes one side of the wave slow down or speed up ...\nLink: https://homework.study.com/explanation/how-is-sound-refracted-going-from-a-less-dense-media-to-a-denser-media.html\n\nSearch result 7\nTitle: Reflection, Refraction, and Diffraction - The Physics Classroom\nContent: Diffraction of water waves is observed in a harbor as waves bend around small boats and are found to disturb the water behind them. The same waves however are ...\nLink: https://www.physicsclassroom.com/class/waves/Lesson-3/Reflection,-Refraction,-and-Diffraction\n\nSearch result 8\nTitle: What happens to a sound wave as it travels from air into water? (a ...\nContent: The sound intensity in the water will be less than it was in air because some sound is reflected by the water surface. However, the frequency ( ...\nLink: https://www.toppr.com/ask/question/what-happens-to-a-sound-wave-as-it-travels-from-air-into-water-a-its/\n\nSearch result 9\nTitle: Ultrasound Physics and Instrumentation - StatPearls - NCBI Bookshelf\nContent: The difference in structure density promotes the refraction or bending of sound waves off the surface. The result is that echoes do not return ...\nLink: https://www.ncbi.nlm.nih.gov/books/NBK570593/\n\nSearch result 10\nTitle: How does sound propagate from air into water?\nContent: A portion of the sound wave will reflect away from the water and into the air, while another part will transmit into the water. During ...\nLink: https://dosits.org/science/movement/how-does-sound-propagate-from-air-into-water/"

Your decision: ###Supported###

Claim: Mount Katahdin is 6,288.2 feet (1,917.6 meters) tall.

Search result 1\nTitle: Mount Katahdin - Wikipedia\nContent: Mount Katahdin is the highest mountain in the U.S. state of Maine at 5,269 feet (1,606 m). Named Katahdin, which means \"Great Mountain\", by the Penobscot ...\nLink: https://en.wikipedia.org/wiki/Mount_Katahdin\n\nSearch result 2\nTitle: Mount Katahdin - Baxter State Park (U.S. National Park Service)\nContent: Katahdin, which translates to \"Greatest Mountain\" in Penobscot, is the highest mountain in the state of Maine at 5,269 feet.\nLink: https://www.nps.gov/places/katahdin-baxter-state-park.htm\n\nSearch result 3\nTitle: Mount Katahdin | Maine, Map, & Facts | Britannica\nContent: Mount Katahdin, highest point (5268 feet [1606 metres]) in Maine, U.S. It lies in Baxter State Park, 20 miles (32 km) northwest of Millinocket, ...\nLink: https://www.britannica.com/place/Mount-Katahdin\n\nSearch result 4\nTitle: Mount Katahdin - Simple English Wikipedia, the free encyclopedia\nContent: It is 5,267 feet (1,605 m) tall. Mount Katahdin. Katahdin from 10,000 ft (3,000 m). Highest point. Elevation, 5,267 ft (1,605 m)NAVD 88 · Prominence, 4,288 ft ( ...\nLink: https://simple.wikipedia.org/wiki/Mount_Katahdin\n\nSearch result 5\nTitle: Mount Katahdin- The Beginning - Alt Route Meals\nContent: It is the states highest peak standing at 5,269ft tall. This is the Northern Terminus of the Appalachian Trail, a 2,189 mile footpath that ...\nLink: https://altroutemeals.com/blogs/news/thai-curry-maine-katahdin-millinocket\n\nSearch result 6\nTitle: Mount Katahdin - PeakVisor\nContent: Mount Katahdin is the highest mountain in the U.S. state of Maine at 5,267 feet (1,605 m). Named Katahdin by the Penobscot Indians, which means \"The ...\nLink: https://peakvisor.com/peak/mount-katahdin.html\n\nSearch result 7\nTitle: Katahdin, Maine - Peakbagger.com\nContent: Elevation: 5268 feet, 1606 meters ; Highest Summit, Baxter Peak ; Subpeaks, Katahdin - South Peak (5260 ft/1603 m) Pamola Peak (4919 ft/1499 m) Chimney Peak (4900 ...\nLink: https://peakbagger.com/Peak.aspx?pid=6820\n\nSearch result 8\nTitle: Summit to Mt. Katahdin - Tallest Peak in Maine. - YouTube\nContent: In 1989, I attempted to summit Mt. Katahdin as a Boy Scout while attending a High Adventure ...\nLink: https://www.youtube.com/watch?v=t-bT2clu57o

Your decision: ###Contradicted###

Your task:
Claim: {{atom_text}}

{{search_results}}

Your decision:
"""

class FactVerify:
    """
    Implementation of the SAFE and subsequent papers. Atomic units are verified using
    Google search results only (without retrieving the context i.e., text).

    Source:
        @misc{wei2024longformfactualitylargelanguage,
            title={Long-form factuality in large language models}, 
            author={Jerry Wei and Chengrun Yang and Xinying Song and Yifeng Lu and Nathan Hu and Jie Huang and Dustin Tran and Daiyi Peng and Ruibo Liu and Da Huang and Cosmo Du and Quoc V. Le},
            year={2024},
            eprint={2403.18802},
            archivePrefix={arXiv},
            primaryClass={cs.CL},
            url={https://arxiv.org/abs/2403.18802}, 
        }    
    """

    def __init__(
            self,
            backend: Backend,
            atom_extractor: Atomizer = None,
            atom_reviser: Reviser = None,
            context_retriever: ContextRetriever = None,
    ):
        """
        Initialize the FactVerify pipeline.

        Args:
            backend: Backend
                The Mellea backend to use for LLM interactions.
            atom_extractor: Atomizer
                The atom decomposition component.
            atom_reviser: Reviser
                The atom reviser component.
            context_retriever: ContextRetriever
                The service used for retrieving external contexts.
        """

        self.backend = backend
        self.query = None
        self.response = None
        self.topic = None
        self.start_time = time.perf_counter() # get the start time

        self.context_retriever = context_retriever
        self.atom_extractor = atom_extractor
        self.atom_reviser = atom_reviser
        self.binary_output = False # we have a 3-label output
       
        print(f"[FactVerify] Using Mellea backend: {self.backend.model_id}")
        print(f"[FactVerify] Binary output: {self.binary_output}")

        self.atoms = {} # indexed by atom id
        self.contexts = {} # indexed by context id

        # Ground truth labels (if any)
        self.labels_human = None

        # Disable Mellea logging
        FancyLogger.get_logger().setLevel(FancyLogger.ERROR)

    def from_dict_with_contexts(
            self,
            data: Dict[str, Any],
    ):
        """
        Initialize FactVerify from a dict containing both atoms and contexts.

        Args:
            data: Dict[str, Any]
                The dict containing the problem instance.
        """

        self.query = data["input"]
        self.response = data["output"]
        
        print(f"[FactVerify] Reading the atoms ...")                
        gold_labels = []
        atom_ids = []
        self.atoms = {}
        self.contexts = {}
        atom2contexts = {}
        for elem in data["atoms"]:
            aid = elem["id"]
            text = elem["text"]
            original = elem["original"]
            label = elem.get("label", None)
            contexts = elem["contexts"]
            a = Atom(id=aid, text=text, label=label)
            a.set_original(original)
            atom_ids.append(aid)
            gold_labels.append(label)
            self.atoms[aid] = a
            atom2contexts[aid] = contexts

        print(f"[FactVerify] Atoms found: {len(self.atoms)}")
        for _, atom in self.atoms.items():
            print(f"[FactVerify] {atom}")

        self.labels_human = dict(zip(atom_ids, gold_labels))
        print(f"[FactVerify] Labels found: {self.labels_human}")

        print(f"[FactVerify] Reading the contexts ...")
        for elem_dict in data["contexts"]:
            cid = elem_dict["id"]
            title = elem_dict["title"]
            text = elem_dict["text"]
            snippet = elem_dict.get("snippet", "")
            link = elem_dict.get("link", "")
            ctxt = Context(
                id=cid, 
                atom=None, 
                text=text, 
                title=title, 
                snippet=snippet, 
                link=link
            )
            self.contexts[cid] = ctxt

        print(f"[FactVerify] Contexts found: {len(self.contexts)}")
        for aid, atom in self.atoms.items():
            ctxts = []
            for c in atom2contexts[aid]:
                ctxts.append(self.contexts[c])
                self.contexts[c].set_atom(atom)
            atom.add_contexts(ctxts)
        print(f"[FactVerify] Pipeline initialized with {len(self.atoms)} atoms and {len(self.contexts)} contexts.")

    def to_json(self, json_file_path: str = None) -> Dict[str, Any]:
        """
        Save the FactVerify instance to a JSON file.

        Args:
            json_file: str
                The path to the output JSON file.
        """

        data = {}
        data["input"] = self.query
        data["output"] = self.response.strip()
        data["topic"] = self.topic
        data["atoms"] = []
        data["contexts"] = []

        # Write the atoms
        for aid, atom in self.atoms.items():
            atom_data = dict(
                id=aid, text=atom.get_text(), contexts=list(atom.get_contexts().keys())
            )
            if atom.get_label() is not None:
                atom_data["label"] = atom.get_label()
            data["atoms"].append(atom_data)

        # Write the contexts
        data["contexts"] = [context.to_json() for context in self.contexts.values()]

        # Write to a JSON file (if any)
        if json_file_path:
            with open(json_file_path, "w") as f:
                json.dump(data, f, indent=4)
            f.close()
            print(f"[FactReasoner] Pipeline instance written to: {json_file_path}")

        return data

    def build(
            self,
            query: str = None,
            response: str = None,
            topic: str = None,
            has_atoms: bool = False,
            has_contexts: bool = False,
            revise_atoms: bool = False,
            use_fast_retriever: bool = True,
    ):
        """
        Build the atoms and contexts using the retrieval service.

        Args:
            query: str
                The input user query.
            response: str
                The LLM generated response to the input query.
            topic: str
                The topic of the input query/response.
            has_atoms: bool
                A boolean flag indicating if the atoms have already been created.
            has_contexts: bool
                A boolean flag indicating if the contexts have already been created.
            revise_atoms: bool
                A boolean flag indicating that the atoms need to be decontextualized
                (i.e., pronouns he, she, it, ... replaced by the actual entity)
            use_fast_retriever: bool
                Use the fast multi-threaded context retriever.
        """

        # Initialize the scorer
        if query is not None: 
            self.query = query
        if response is not None:
            self.response = response
        if topic is not None:
            self.topic = topic

        self.revise_atoms = revise_atoms

        # Safety checks
        assert self.atom_extractor is not None, \
            f"The atom extractor must be created."
        assert self.atom_reviser is not None, \
            f"The atom reviser must be created."

        print(f"[FactVerify] Building the pipeline ...")
        
        # Build the atoms 
        if has_atoms == False:
            self.atoms = build_atoms(
                response=self.response,
                atom_extractor=self.atom_extractor
            )
            self.revise_atoms = True # revise atoms if newly created
            print(f"[FactVerify] Extracted {len(self.atoms)} atoms.")
            for aid in self.atoms.keys():
                print(f"[FactVerify] {self.atoms[aid]}")

        assert len(self.atoms) > 0, \
            f"The atoms must be initialized before running the pipeline."

        # Revise the atoms
        if self.revise_atoms:
            print(f"[FactVerify] Revise the atoms ...")
            assert self.response is not None, f"The atom reviser requires a response."
            atom_ids = [aid for aid in sorted(self.atoms.keys())]
            old_atoms = [self.atoms[aid].get_text() for aid in atom_ids]
            result = asyncio.run(self.atom_reviser.run_batch(old_atoms, self.response))
            for i, aid in enumerate(atom_ids):
                elem = result[i]
                self.atoms[aid].set_text(elem["revised_unit"])
                print(f"[FactVerify] {self.atoms[aid]}")

        # Remove duplicated atoms (if any)
        self.atoms = remove_duplicated_atoms(self.atoms)
        print(f"[FactVerify] Created {len(self.atoms)} unique atoms.")

        # Build the contexts (per atom)
        if has_contexts == False: # check if contexts already in file
            self.contexts = build_contexts(
                atoms=self.atoms,
                retriever=self.context_retriever,
                use_fast_retriever=use_fast_retriever,
            )

    def _get_label(self, output: ModelOutputThunk) -> str:
        """
        Extract the atom label from the generated text. We expect the label to
        be on the last line of the response, and be one of the following:
            [Supported], [Contradicted], [Inconclusive].
        
        Depending on the binary_output flag we return:
            - [Supported]/S atoms, the others will be [NotSupported]/NS.
            - [Supported]/S atoms, [Contradicted]/C atoms and [Inconclusive]/U atoms.
        """

        text = str(output)
        label = extract_last_wrapped_response(text)
        if self.binary_output:
            if len(label) > 0 and label.lower() in ['supported']:
                return "S"
            else:
                return "NS"
        else:
            if len(label) > 0 and label.lower() in ['supported']:
                return "S"
            elif len(label) > 0 and label.lower() in ['contradicted']:
                return "C"
            else:
                return "U"
       
    async def predict_atom_labels(self) -> Tuple[Dict[str, str], Dict[str, str]]:
        """
        For each atom, predict its label given the corresponding retrieved contexts.
        """

        # Safety checks
        assert len(self.atoms) > 0

        # Utility function to assemble the context of an atom
        def make_search_results(search_results: List[Dict[str, Any]]) -> str:
            i = 1
            search_results_str = ""
            for dict_item in search_results:
                title = dict_item["title"]
                snippet = dict_item["snippet"]
                link = dict_item["link"]
                search_results_str += f"Search Result {i}\nTitle: {title}\nContent: {snippet}\nLink: {link}\n\n"
                i += 1
            return search_results_str

        # Use the LLM to label the atom
        print(f"[FactVerify] Labeling atoms with {self.backend.model_id} ...")

        # Create the prompts for each of the atoms
        atom_ids = []
        atom_labels = []
        atom_outputs = []
        corutines = []
        for aid, atom in self.atoms.items():
            atom_ids.append(aid)
            atom_text = atom.get_text()
            contexts = atom.get_contexts()
            search_results = []
            if contexts is not None and len(contexts) > 0:
                search_results = [dict(title=c.get_title(), snippet=c.get_snippet(), link=c.get_link()) for _, c in contexts.items()]

            # Prepare the context
            search_results_text = make_search_results(search_results)
            print(f"[FactVerify] Processing atom: ({aid}) {atom_text}")

            # Execute the instruction
            corutine = mfuncs.ainstruct(
                INSTRUCTION_FACTVERIFY,
                context=SimpleContext(),
                backend=self.backend,
                requirements=[
                    check(
                        "The output must contain [Supported], [Contradicted] or [Inconclusive]"
                    )
                ],
                user_variables={"atom_text": atom_text, "search_results": search_results_text},
                strategy=RejectionSamplingStrategy(loop_budget=LOOP_BUDGET),
                return_sampling_results=True
            )
            corutines.append(corutine)
        
        print(f"[FactVerify] Awaiting for the async execution ...")
        outputs = await asyncio.gather(*(corutines[i] for i in range(len(corutines))))
        for output in outputs:
            label = self._get_label(output.result)
            atom_labels.append(label)
            atom_outputs.append(str(output))

        # Return the labeled atoms
        return dict(zip(atom_ids, atom_labels)), dict(zip(atom_ids, atom_outputs))
    
    def score(self):
        """
        Compute the factuality score taking into consideration the contexts 
        retrieved for each of the atom in the answer.

        Factuality score = # atoms(true) / # atoms

        Intuitively, a score of 100% means that all atoms in the answer are
        factually correct. If none of them are correct, then the score is 0%. If
        only half of the atoms are correct, then the score is 50%.

        Returns:
            Dict[str, Any]: The results dictionary containing the factuality score i.e., a real value in [0, 1]
        """

        # Compute the factuality score (i.e., precision)
        num_true_atoms = 0
        num_false_atoms = 0
        num_uniform_atoms = 0
        labels, raw_outputs = asyncio.run(self.predict_atom_labels())
        for _, label in labels.items():
            if self.binary_output:
                if label == "S":
                    num_true_atoms += 1
                else:
                    num_false_atoms += 1
            else:
                if label == "S":
                    num_true_atoms += 1
                elif label == "C":
                    num_false_atoms += 1
                else:
                    num_uniform_atoms += 1
        
        # Precision, R@K and F1@K
        fscore = float(num_true_atoms)/float(len(self.atoms))
        K = int(len(self.atoms) / 2) # K is assumed to be half
        recall_k = min(float(num_true_atoms)/K, 1)
        try:
            f1k = 2 * fscore * recall_k / (fscore + recall_k)
        except Exception as _:
            f1k = 0.0

        # Elapsed time
        elapsed_time = time.perf_counter() - self.start_time # elapsed time

        results = {}
        results["factuality_score"] = fscore
        results["recall_k"] = recall_k
        results["f1_k"] = f1k
        results["num_atoms"] = len(self.atoms)
        results["num_contexts"] = len(self.contexts)
        results["num_true_atoms"] = num_true_atoms
        results["num_false_atoms"] = num_false_atoms
        results["num_uniform_atoms"] = num_uniform_atoms
        results["entropy"] = None
        results["avg_entropy"] = None

        print(f"[FactVerify] Predictions: {labels}")
        if self.labels_human is not None and self.binary_output is True:
            true_atoms = 0
            false_atoms = 0
            num_true_positive = 0
            num_true_negative = 0
            num_false_positive = 0
            num_false_negative = 0
            for aid, l in self.labels_human.items():
                if l == "S":
                    true_atoms += 1
                    if labels[aid] == "S":
                        num_true_positive += 1
                    else:
                        num_false_negative += 1
                else:
                    false_atoms += 1
                    if labels[aid] == "NS":
                        num_true_negative += 1
                    else:
                        num_false_positive += 1     

            fscore_gold = true_atoms/len(self.labels_human)
            print(f"[FactVerify] Gold labels: {self.labels_human}")
            print(f"[FactVerify] Predictions: {labels}")
            print(f"[FactVerify] Gold fscore: {fscore_gold} ({true_atoms}/{len(self.labels_human)})")
            results["gold_factuality_score"] = fscore_gold
            results["gold_true_atoms"] = true_atoms
            results["true_positive"] = num_true_positive
            results["true_negative"] = num_true_negative
            results["false_positive"] = num_false_positive
            results["false_negative"] = num_false_negative
        elif self.labels_human is not None and self.binary_output is False:
            true_atoms = 0
            false_atoms = 0
            num_true_positive = 0
            num_true_negative = 0
            num_false_positive = 0
            num_false_negative = 0
            for aid, l in self.labels_human.items(): # true labels are either S or NS
                if l == "S":
                    true_atoms += 1
                    if labels[aid] == "S":
                        num_true_positive += 1
                    else:
                        num_false_negative += 1
                else:
                    false_atoms += 1
                    if labels[aid] in ["C", "U"]:
                        num_true_negative += 1
                    else:
                        num_false_positive += 1     

            fscore_gold = true_atoms/len(self.labels_human)
            print(f"[FactVerify] Gold labels: {self.labels_human}")
            print(f"[FactVerify] Predictions: {labels}")
            print(f"[FactVerify] Gold fscore: {fscore_gold} ({true_atoms}/{len(self.labels_human)})")
            results["gold_factuality_score"] = fscore_gold
            results["gold_true_atoms"] = true_atoms
            results["true_positive"] = num_true_positive
            results["true_negative"] = num_true_negative
            results["false_positive"] = num_false_positive
            results["false_negative"] = num_false_negative

        results["topic"] = self.topic
        results["query"] = self.query
        results["response"] = self.response
        results["elapsed_time"] = elapsed_time
        results["predictions"] = labels
        results["raw_outputs"] = raw_outputs
        print(f"[FactVerify] Elapsed time: {elapsed_time:.4f} seconds.")

        return results




"""FactReasoner - a probabilistic factuality assessor for LLMs"""

from .assessor import FactReasoner
from .corrector import FactCorrector
from .fact_graph import FactGraph
from .search_api import SearchAPI
from .utils import (
    extract_first_square_brackets,
    extract_last_square_brackets,
    extract_first_code_block,
    extract_last_wrapped_response,
    strip_code_fences,
    strip_string,
    normalize_ws,
    validate_json_code_block,
    validate_markdown_code_block,
)

__all__ = [
    "FactReasoner",
    "FactCorrector",
    "FactGraph",
    "SearchAPI",
    "extract_first_square_brackets",
    "extract_last_square_brackets",
    "extract_first_code_block",
    "extract_last_wrapped_response",
    "strip_code_fences",
    "strip_string",
    "normalize_ws",
    "validate_json_code_block",
    "validate_markdown_code_block",
]
